/*
 * Copyright (C) 2013 Kamil Dudka <kdudka@redhat.com>
 *
 * This file is part of predator.
 *
 * predator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * predator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with predator.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"
#include "adt_op_def.hh"

#include "adt_op.hh"
#include "symtrace.hh"

#include <cl/storage.hh>

namespace AdtOp {

class TplFactory {
    public:
        TplFactory(const CodeStorage::Storage &stor):
            stor_(stor),
            node_(new Trace::TransientNode("TplFactory")),
            ptrSize_(stor.types.dataPtrSizeof()),
            objSize_(IR::rngFromNum(2 * ptrSize_))
        {
            bOff_.head = 0;
            bOff_.next = 0;
            bOff_.prev = ptrSize_;
        }

        TOffset headAt() const {
            return bOff_.head;
        }

        TOffset nextAt() const {
            return bOff_.next;
        }

        TOffset prevAt() const {
            return bOff_.prev;
        }

        SymHeap createHeap() const {
            return SymHeap(stor_, node_.node());
        }

        TObjId createObj(SymHeap *pSh, EObjKind kind) const;

        void nullFieldsOfObj(SymHeap *pSh, TObjId obj) const;

        void dropFieldsOfObj(SymHeap *pSh, TObjId obj) const;

    private:
        const CodeStorage::Storage         &stor_;
        const Trace::NodeHandle             node_;
        const TSizeOf                       ptrSize_;
        const TSizeRange                    objSize_;
        BindingOff                          bOff_;
};

TObjId TplFactory::createObj(SymHeap *pSh, const EObjKind kind) const
{
    const TObjId obj = pSh->heapAlloc(objSize_);
    if (OK_REGION != kind)
        pSh->objSetAbstract(obj, kind, bOff_);

    return obj;
}

void TplFactory::nullFieldsOfObj(SymHeap *pSh, const TObjId obj) const
{
    // obtain handles for next/prev fields
    const PtrHandle nextPtr(*pSh, obj, this->nextAt());
    const PtrHandle prevPtr(*pSh, obj, this->prevAt());

    // nullify the next/prev fields
    nextPtr.setValue(VAL_NULL);
    prevPtr.setValue(VAL_NULL);
}

void TplFactory::dropFieldsOfObj(SymHeap *pSh, const TObjId obj) const
{
    const TValId valUnknown = pSh->valCreate(VT_UNKNOWN, VO_ASSIGNED);

    const UniformBlock ubAll = {
        /* off      */  0,
        /* size     */  objSize_.lo,
        /* tplValue */  valUnknown
    };

    pSh->writeUniformBlock(obj, ubAll);
}

enum EListSide {
    LS_INVALID,
    LS_FRONT,
    LS_BACK
};

void connectPush(
        SymHeap                    &sh,
        const TplFactory           &fact,
        const TObjId                dls,
        const TObjId                reg,
        const EListSide             side)
{
    TOffset offNext;
    TOffset offPrev;
    ETargetSpecifier ts;
    switch (side) {
        case LS_FRONT:
            offNext = fact.prevAt();
            offPrev = fact.nextAt();
            ts = TS_FIRST;
            break;

        case LS_BACK:
            offNext = fact.nextAt();
            offPrev = fact.prevAt();
            ts = TS_LAST;
            break;

        default:
            CL_BREAK_IF("invalid call of connectPush()");
            return;
    }

    // obtain handles for next/prev fields
    const PtrHandle nextPtr(sh, reg, offNext);
    const PtrHandle prevPtr(sh, reg, offPrev);
    const PtrHandle endPtr (sh, dls, offNext);

    // chain both objects together such that they represent a linked list
    const TValId regAt = sh.addrOfTarget(reg, TS_REGION, fact.headAt());
    const TValId endAt = sh.addrOfTarget(dls, ts,        fact.headAt());
    endPtr.setValue(regAt);
    prevPtr.setValue(endAt);
    nextPtr.setValue(VAL_NULL);
}

OpTemplate* createPushByRef(TplFactory &fact, const EListSide side)
{
    OpTemplate *tpl = new OpTemplate((side == LS_FRONT)
            ? "push_front_by_ref"
            : "push_back_by_ref", SHAPE_DLL_NULL);

    SymHeap sh(fact.createHeap());

    // allocate an uninitialized region
    const TObjId reg = fact.createObj(&sh, OK_REGION);
    SymHeap input(sh);
    Trace::waiveCloneOperation(input);

    // nullify the next/prev fields
    fact.nullFieldsOfObj(&sh, reg);

    // register pre/post pair for push_back() to an empty list
    SymHeap output(sh);
    Trace::waiveCloneOperation(output);
    OpFootprint *fp = new OpFootprint(input, output);
    fp->inArgs.push_back(reg);
    tpl->addFootprint(fp);

    // drop the nullified fields of 'reg'
    fact.dropFieldsOfObj(&sh, reg);

    // allocate a DLS that will represent a container shape in our template
    const TObjId dls = fact.createObj(&sh, OK_DLS);
    fact.nullFieldsOfObj(&sh, dls);
    input = sh;

    // chain both objects together such that they represent a linked list
    connectPush(sh, fact, dls, reg, side);
    output = sh;

    // register pre/post pair for push_back() to a non-empty list
    Trace::waiveCloneOperation(input);
    Trace::waiveCloneOperation(output);
    fp = new OpFootprint(input, output);
    fp->inArgs.push_back(reg);
    tpl->addFootprint(fp);

    return tpl;
}

OpTemplate* createPushByVal(TplFactory &fact, const EListSide side)
{
    OpTemplate *tpl = new OpTemplate((side == LS_FRONT)
            ? "push_front_by_val"
            : "push_back_by_val", SHAPE_DLL_NULL);

    SymHeap sh(fact.createHeap());
    SymHeap input(sh);
    Trace::waiveCloneOperation(input);

    // allocate an uninitialized region
    TObjId reg = fact.createObj(&sh, OK_REGION);

    // nullify the next/prev fields
    fact.nullFieldsOfObj(&sh, reg);

    // register pre/post pair for push_back() to an empty list
    SymHeap output(sh);
    Trace::waiveCloneOperation(output);
    OpFootprint *fp = new OpFootprint(input, output);
    fp->outArgs.push_back(reg);
    tpl->addFootprint(fp);

    // start with a fresh heap
    sh = fact.createHeap();
    Trace::waiveCloneOperation(sh);

    // allocate a DLS that will represent a container shape in our template
    const TObjId dls = fact.createObj(&sh, OK_DLS);
    fact.nullFieldsOfObj(&sh, dls);
    input = sh;

    // allocate an uninitialized region
    reg = fact.createObj(&sh, OK_REGION);

    // chain both objects together such that they represent a linked list
    connectPush(sh, fact, dls, reg, side);
    output = sh;

    // register pre/post pair for push_back() to a non-empty list
    Trace::waiveCloneOperation(input);
    Trace::waiveCloneOperation(output);
    fp = new OpFootprint(input, output);
    fp->outArgs.push_back(reg);
    tpl->addFootprint(fp);

    return tpl;
}

// cyclic DLL without head
OpTemplate* createCDLLPushByVal(TplFactory &fact, const EListSide side)
{
    TOffset offNext;
    TOffset offPrev;
    ETargetSpecifier ts;
    OpTemplate *tpl = new OpTemplate((side == LS_FRONT) ?
                          "cdll_push_front_by_val" : "cdll_push_back_by_val",
                           SHAPE_CYCLIC_DLL);

    switch (side) {
        case LS_FRONT:
            offNext = fact.prevAt();
            offPrev = fact.nextAt();
            ts = TS_FIRST;
            break;

        case LS_BACK:
            offNext = fact.nextAt();
            offPrev = fact.prevAt();
            ts = TS_LAST;
            break;

        default:
            CL_BREAK_IF("invalid call of createCDLLPushByVal()");
            return new OpTemplate("createCDLLPushByVal_invalid", SHAPE_UNKNOWN);
    }

// variant1) add to empty CDLL: 0+1->1
#if 1   // OK
{
    SymHeap sh(fact.createHeap());
    SymHeap input(sh);                  // create empty INPUT heap
    Trace::waiveCloneOperation(input);

    // allocate an uninitialized region
    TObjId reg1 = fact.createObj(&sh, OK_REGION);

    // link to itself
    const PtrHandle reg1Next(sh, reg1, offNext);
    const PtrHandle reg1Prev(sh, reg1, offPrev);
    const TValId reg1At = sh.addrOfTarget(reg1, TS_REGION, fact.headAt());
    reg1Prev.setValue(reg1At);
    reg1Next.setValue(reg1At);

    // register pre/post pair for push_back() to an empty list
    SymHeap output(sh);                 // create OUTPUT heap with single item cDLL
    Trace::waiveCloneOperation(output);

    // register variant empty->[1]
    OpFootprint *fp = new OpFootprint(input, output);
    fp->outArgs.push_back(reg1);
    tpl->addFootprint(fp);
}
#endif

#if 0   // OK
// variant2) 1+1->2
{
    SymHeap sh(fact.createHeap());
    // start with a single item list

    // allocate an uninitialized region
    TObjId reg1 = fact.createObj(&sh, OK_REGION);
    // nullify the next/prev fields
//    fact.nullFieldsOfObj(&sh, reg1);

    // link to itself
    const PtrHandle reg1Next(sh, reg1, offNext);
    const PtrHandle reg1Prev(sh, reg1, offPrev);
    const TValId reg1At = sh.addrOfTarget(reg1, TS_REGION, fact.headAt());
    reg1Prev.setValue(reg1At);
    reg1Next.setValue(reg1At);

    SymHeap input(sh);                  // create single item INPUT heap
    Trace::waiveCloneOperation(input);

    // allocate an uninitialized region
    TObjId reg2 = fact.createObj(&sh, OK_REGION);
    // nullify the next/prev fields
//    fact.nullFieldsOfObj(&sh, reg2);

    // chain both regions together such that they represent a cDLL
    const PtrHandle reg2Next(sh, reg2, offNext);
    const PtrHandle reg2Prev(sh, reg2, offPrev);
    const TValId reg2At = sh.addrOfTarget(reg2, TS_REGION, fact.headAt());
    reg1Prev.setValue(reg2At);
    reg1Next.setValue(reg2At);
    reg2Prev.setValue(reg1At);
    reg2Next.setValue(reg1At);

    // register pre/post pair for push_back() to an empty list
    SymHeap output(sh);                 // create OUTPUT heap with 2 item cDLL
    Trace::waiveCloneOperation(output);

    // register variant [1]+1->2
    OpFootprint *fp = new OpFootprint(input, output);
    fp->outArgs.push_back(reg2);
    tpl->addFootprint(fp);
}
#endif

#if 0
// variant2a) cyclic((REG+REG)+REG)
{
    SymHeap sh(fact.createHeap());
    // start with a single item list

    // allocate an uninitialized region
    TObjId reg1 = fact.createObj(&sh, OK_REGION);
    const PtrHandle reg1Next(sh, reg1, offNext);
    const PtrHandle reg1Prev(sh, reg1, offPrev);
    const TValId reg1At = sh.addrOfTarget(reg1, TS_REGION, fact.headAt());
    // allocate an uninitialized region
    TObjId reg2 = fact.createObj(&sh, OK_REGION);
    const PtrHandle reg2Next(sh, reg2, offNext);
    const PtrHandle reg2Prev(sh, reg2, offPrev);
    const TValId reg2At = sh.addrOfTarget(reg2, TS_REGION, fact.headAt());
    // chain both regions together such that they represent a cDLL
    reg1Prev.setValue(reg2At);
    reg1Next.setValue(reg2At);
    reg2Prev.setValue(reg1At);
    reg2Next.setValue(reg1At);

    SymHeap input(sh);                  // create single item INPUT heap
    Trace::waiveCloneOperation(input);

    // allocate an uninitialized region
    TObjId reg3 = fact.createObj(&sh, OK_REGION);
    const PtrHandle reg3Next(sh, reg3, offNext);
    const PtrHandle reg3Prev(sh, reg3, offPrev);
    const TValId reg3At = sh.addrOfTarget(reg3, TS_REGION, fact.headAt());
    // chain regions together such that they represent a cDLL
    reg1Prev.setValue(reg3At);
    reg2Next.setValue(reg3At);
    reg3Prev.setValue(reg2At);
    reg3Next.setValue(reg1At);

    // register pre/post pair for push_back() to an empty list
    SymHeap output(sh);                 // create OUTPUT heap with 2 item cDLL
    Trace::waiveCloneOperation(output);

    // register variant [1]+1->2
    OpFootprint *fp = new OpFootprint(input, output);
    fp->outArgs.push_back(reg2);
    tpl->addFootprint(fp);
}
#endif

#if 0   // NOT
// variant3a)   create cyclic (DLS+REG)+REG
{
    SymHeap sh(fact.createHeap());
    // start with a fresh heap

    // allocate an uninitialized region
    TObjId reg0 = fact.createObj(&sh, OK_REGION);
    const PtrHandle reg0Next(sh, reg0, offNext);
    const PtrHandle reg0Prev(sh, reg0, offPrev);
    const TValId reg0At = sh.addrOfTarget(reg0, TS_REGION, fact.headAt());

    // allocate a DLS that will represent a container shape in our template
    const TObjId dls = fact.createObj(&sh, OK_DLS);
    const PtrHandle dlsNext(sh, dls, offNext);
    const PtrHandle dlsPrev(sh, dls, offPrev);
    const TValId dlsAtLast = sh.addrOfTarget(dls, ts, fact.headAt());
    const TValId dlsAtFirst = sh.addrOfTarget(dls, (ts==TS_LAST)?TS_FIRST:TS_LAST, fact.headAt());

    // chain both objects together such that they represent a cDLL
    dlsNext.setValue(reg0At);
    dlsPrev.setValue(reg0At);
    reg0Prev.setValue(dlsAtLast);
    reg0Next.setValue(dlsAtFirst);

    SymHeap input(sh);                  // create INPUT heap c(DLS+REG)
    Trace::waiveCloneOperation(input);
    /////////////////////////////////////////////////////////////////////////////

    // allocate an uninitialized region
    TObjId reg = fact.createObj(&sh, OK_REGION);
    const PtrHandle regNext(sh, reg, offNext);
    const PtrHandle regPrev(sh, reg, offPrev);
    const TValId regAt = sh.addrOfTarget(reg, TS_REGION, fact.headAt());

    // chain objects together such that they represent a cDLL
    dlsNext.setValue(regAt);
    reg0Prev.setValue(regAt);
    regPrev.setValue(dlsAtLast);
    regNext.setValue(reg0At);

    // register pre/post pair for push_back() to an empty list
    SymHeap output(sh);                 // create OUTPUT heap c(DLS+REG+REG)
    Trace::waiveCloneOperation(output);

    /////////////////////////////////////////////////////////////////////////////
    // register variant [1]+1->2
    OpFootprint *fp = new OpFootprint(input, output);
    fp->outArgs.push_back(reg);
    tpl->addFootprint(fp);
}
#endif

#if 1   //
// variant3)  SIMPLE  cyclic((DLS)+REG)
{
    SymHeap sh(fact.createHeap());
    // start with a single item list

    // allocate a DLS that will represent a container shape in our template
    const TObjId dls = fact.createObj(&sh, OK_DLS);
    const PtrHandle dlsNext(sh, dls, offNext);
    const PtrHandle dlsPrev(sh, dls, offPrev);
    const TValId dlsAtLast = sh.addrOfTarget(dls, ts, fact.headAt());
    const TValId dlsAtFirst = sh.addrOfTarget(dls, (ts==TS_LAST)?TS_FIRST:TS_LAST, fact.headAt());
    // link DLS to itself
    dlsPrev.setValue(dlsAtLast);
    dlsNext.setValue(dlsAtFirst);

    SymHeap input(sh);                  // create INPUT heap cDLS + REG
    Trace::waiveCloneOperation(input);

    // allocate an uninitialized region
    TObjId reg = fact.createObj(&sh, OK_REGION);
    const PtrHandle regNext(sh, reg, offNext);
    const PtrHandle regPrev(sh, reg, offPrev);
    const TValId regAt = sh.addrOfTarget(reg, TS_REGION, fact.headAt());

    // chain both objects together such that they represent a cDLL
    dlsNext.setValue(regAt);
    dlsPrev.setValue(regAt);
    regNext.setValue(dlsAtFirst);
    regPrev.setValue(dlsAtLast);

    // register pre/post pair for push_back() to an empty list
    SymHeap output(sh);                 // create OUTPUT heap with 2 item cDLL
    Trace::waiveCloneOperation(output);

    // register variant [1]+1->2
    OpFootprint *fp = new OpFootprint(input, output);
    fp->outArgs.push_back(reg);
    tpl->addFootprint(fp);
}
#endif

    return tpl;
}

/// remove item from cyclic DLL without head
OpTemplate* createCDLLPop(TplFactory &fact, const EListSide side)
{
    TOffset offNext;
    TOffset offPrev;
    ETargetSpecifier ts;
    OpTemplate *tpl;

    switch (side) {
        case LS_FRONT:
            offNext = fact.prevAt();
            offPrev = fact.nextAt();
            ts = TS_FIRST;
            tpl = new OpTemplate("cdll_pop_front", SHAPE_CYCLIC_DLL);
            break;

        case LS_BACK:
            offNext = fact.nextAt();
            offPrev = fact.prevAt();
            ts = TS_LAST;
            tpl = new OpTemplate("cdll_pop_back", SHAPE_CYCLIC_DLL);
            break;

        default:
            CL_BREAK_IF("invalid call of createCDLLPop()");
            return new OpTemplate("cdll_pop_invalid", SHAPE_UNKNOWN);
    }

#if 1 // OK variant1) 1->0  (remove last item)
{
    SymHeap sh(fact.createHeap());

    // allocate an uninitialized region
    TObjId reg1 = fact.createObj(&sh, OK_REGION);

    // link to itself
    const PtrHandle reg1Next(sh, reg1, offNext);
    const PtrHandle reg1Prev(sh, reg1, offPrev);
    const TValId reg1At = sh.addrOfTarget(reg1, TS_REGION, fact.headAt());
    reg1Prev.setValue(reg1At);
    reg1Next.setValue(reg1At);

    // store the input heap
    SymHeap input(sh);
    Trace::waiveCloneOperation(input);

    // free the region
    sh.objInvalidate(reg1);

    // register pre/post pair for pop_back() on a singleton list
    // sh is OUTPUT heap with region invalidated
    OpFootprint *fp = new OpFootprint(input, /* output */ sh);
//    fp->outArgs.push_back(reg1);        // what region was removed
    tpl->addFootprint(fp);
}
#endif

#if 1  // STANDARD  N->N-1 for N>=2
// variant2)  remove CDLL item   cyclic(DLS+REG)->cyclic(DLS)
{
    SymHeap sh(fact.createHeap());
    // start with a c(DLS+REG) list

    // allocate a DLS that will represent a container shape in our template
    // TODO: ensure DLS1+ ?
    const TObjId dls = fact.createObj(&sh, OK_DLS);
    const PtrHandle dlsNext(sh, dls, offNext);
    const PtrHandle dlsPrev(sh, dls, offPrev);
    const TValId dlsAtLast = sh.addrOfTarget(dls, ts, fact.headAt());
    const TValId dlsAtFirst = sh.addrOfTarget(dls, (ts==TS_LAST)?TS_FIRST:TS_LAST, fact.headAt());

    // allocate an uninitialized region
    TObjId reg = fact.createObj(&sh, OK_REGION);
    const PtrHandle regNext(sh, reg, offNext);
    const PtrHandle regPrev(sh, reg, offPrev);
    const TValId regAt = sh.addrOfTarget(reg, TS_REGION, fact.headAt());

    // chain both objects together such that they represent a cDLL
    dlsNext.setValue(regAt);
    dlsPrev.setValue(regAt);
    regNext.setValue(dlsAtFirst);
    regPrev.setValue(dlsAtLast);

    SymHeap input(sh);                  // create INPUT heap DLS+REG
    Trace::waiveCloneOperation(input);

    // free the region
    sh.objInvalidate(reg);
    // link DLS to itself (should be DLS1+)
    dlsPrev.setValue(dlsAtLast);
    dlsNext.setValue(dlsAtFirst);

    // register pre/post pair for pop_back()
    // sh is OUTPUT heap cDLS
    // register variant 2->1
    OpFootprint *fp = new OpFootprint(input, /* output */ sh);
//    fp->outArgs.push_back(reg);         // what region was removed
    tpl->addFootprint(fp);
}
#endif

#if 0  // EXTRA EXPERIMENT  N->N-1 for N>=2
// variant3)  remove CDLL item   cyclic(REGh+DLS+REG)->cyclic(REGh+DLS)
{
    SymHeap sh(fact.createHeap());
    // start with a c(REG+DLS+REG) list

    // allocate an uninitialized region
    TObjId reg0 = fact.createObj(&sh, OK_REGION);
    const PtrHandle reg0Next(sh, reg0, offNext);
    const PtrHandle reg0Prev(sh, reg0, offPrev);
    const TValId reg0At = sh.addrOfTarget(reg0, TS_REGION, fact.headAt());

    // allocate a DLS
    const TObjId dls = fact.createObj(&sh, OK_DLS);
    const PtrHandle dlsNext(sh, dls, offNext);
    const PtrHandle dlsPrev(sh, dls, offPrev);
    const TValId dlsAtLast = sh.addrOfTarget(dls, ts, fact.headAt());
    const TValId dlsAtFirst = sh.addrOfTarget(dls, (ts==TS_LAST)?TS_FIRST:TS_LAST, fact.headAt());

    // allocate an uninitialized region
    TObjId reg = fact.createObj(&sh, OK_REGION);
    const PtrHandle regNext(sh, reg, offNext);
    const PtrHandle regPrev(sh, reg, offPrev);
    const TValId regAt = sh.addrOfTarget(reg, TS_REGION, fact.headAt());

    // chain all objects together such that they represent a cDLL
    // -> reg0 <-> dls <-> reg <-
    dlsNext.setValue(regAt);
    dlsPrev.setValue(reg0At);
    reg0Next.setValue(dlsAtFirst);
    reg0Prev.setValue(regAt);
    regNext.setValue(reg0At);
    regPrev.setValue(dlsAtLast);

    SymHeap input(sh);                  // create INPUT heap DLS+REG
    Trace::waiveCloneOperation(input);

    // free the region
    sh.objInvalidate(reg);
    // link remaining REG+DLS
    // -> reg0 <-> dls <-
    dlsNext.setValue(reg0At);
    reg0Prev.setValue(dlsAtLast);

    // register pre/post pair for pop_back()
    // sh is OUTPUT heap cDLS
    OpFootprint *fp = new OpFootprint(input, /* output */ sh);
//    fp->inArgs.push_back(reg);         // what region was removed
    tpl->addFootprint(fp);
}
#endif

    return tpl;
}

OpTemplate* createPop(TplFactory &fact, const EListSide side)
{
    TOffset offNext;
    TOffset offPrev;
    ETargetSpecifier ts;
    OpTemplate *tpl;

    switch (side) {
        case LS_FRONT:
            offNext = fact.prevAt();
            offPrev = fact.nextAt();
            ts = TS_FIRST;
            tpl = new OpTemplate("pop_front", SHAPE_DLL_NULL);
            break;

        case LS_BACK:
            offNext = fact.nextAt();
            offPrev = fact.prevAt();
            ts = TS_LAST;
            tpl = new OpTemplate("pop_back", SHAPE_DLL_NULL);
            break;

        default:
            CL_BREAK_IF("invalid call of createPop()");
            return new OpTemplate("pop_invalid", SHAPE_UNKNOWN);
    }

    // allocate a region
    SymHeap sh(fact.createHeap());
    const TObjId reg = fact.createObj(&sh, OK_REGION);

    // nullify the next/prev fields
    const PtrHandle regNext(sh, reg, offNext);
    const PtrHandle regPrev(sh, reg, offPrev);
    regNext.setValue(VAL_NULL);
    regPrev.setValue(VAL_NULL);

    // store the input heap
    SymHeap input(sh);
    Trace::waiveCloneOperation(input);

    // free the region
    sh.objInvalidate(reg);

    // register pre/post pair for pop_back() on a singleton list
    tpl->addFootprint(new OpFootprint(input, /* output */ sh));

    // roll-back the "free" operation
    sh.swap(input);

    // allocate a DLS that will represent a container shape in our template
    const TObjId dls = fact.createObj(&sh, OK_DLS);
    const PtrHandle dlsNext(sh, dls, offNext);
    const PtrHandle dlsPrev(sh, dls, offPrev);

    // chain both objects together such that they represent a linked list
    const TValId regAt = sh.addrOfTarget(reg, TS_REGION, fact.headAt());
    const TValId endAt = sh.addrOfTarget(dls, ts,        fact.headAt());
    dlsPrev.setValue(VAL_NULL);
    regPrev.setValue(endAt);
    dlsNext.setValue(regAt);

    // store the input heap
    input = sh;
    Trace::waiveCloneOperation(input);

    // free the region and re-terminate the remainder of the list
    sh.objInvalidate(reg);
    dlsNext.setValue(VAL_NULL);

    // register pre/post pair for pop_back() on a 2+ list
    tpl->addFootprint(new OpFootprint(input, /* output */ sh));

    return tpl;
}

OpTemplate* createPopRef(TplFactory &fact, const EListSide side)
{
    TOffset offNext;
    TOffset offPrev;
    ETargetSpecifier ts;
    OpTemplate *tpl;

    switch (side) {
        case LS_FRONT:
            offNext = fact.prevAt();
            offPrev = fact.nextAt();
            ts = TS_FIRST;
            tpl = new OpTemplate("pop_front_by_ref", SHAPE_DLL_NULL);
            break;

        case LS_BACK:
            offNext = fact.nextAt();
            offPrev = fact.prevAt();
            ts = TS_LAST;
            tpl = new OpTemplate("pop_back_by_ref", SHAPE_DLL_NULL);
            break;

        default:
            CL_BREAK_IF("invalid call of createPopRef()");
            return new OpTemplate("pop_by_ref_invalid", SHAPE_UNKNOWN);
    }

#if 0
// VARIANT 1) single item list = problem
{
    SymHeap sh(fact.createHeap());
    // allocate a region
    const TObjId reg = fact.createObj(&sh, OK_REGION);

    // nullify the next/prev fields
    const PtrHandle regNext(sh, reg, offNext);
    const PtrHandle regPrev(sh, reg, offPrev);
    regNext.setValue(VAL_NULL);
    regPrev.setValue(VAL_NULL);

    // store the input heap
    SymHeap input(sh);
    Trace::waiveCloneOperation(input);

    // free the region === BUG: for *by_ref should be intact
    sh.objInvalidate(reg);

    // register pre/post pair for pop_back_by_ref() on singleton list
    OpFootprint *fp = new OpFootprint(input, /* output */ sh);
//    fp->outArgs.push_back(reg); // does not make sense if destroyed
    tpl->addFootprint(fp);
}
#endif

// VARIANT 2) DLS+R->DLS
{
    SymHeap sh(fact.createHeap());
    // allocate a DLS that will represent a container shape in our template
    const TObjId dls = fact.createObj(&sh, OK_DLS);
    const PtrHandle dlsNext(sh, dls, offNext);
    const PtrHandle dlsPrev(sh, dls, offPrev);

    // allocate a region
    const TObjId reg = fact.createObj(&sh, OK_REGION);
    const PtrHandle regNext(sh, reg, offNext);
    const PtrHandle regPrev(sh, reg, offPrev);

    // chain both objects together such that they represent a linked list
    const TValId regAt = sh.addrOfTarget(reg, TS_REGION, fact.headAt());
    const TValId endAt = sh.addrOfTarget(dls, ts,        fact.headAt());
    dlsPrev.setValue(VAL_NULL);
    dlsNext.setValue(regAt);
    regNext.setValue(VAL_NULL);
    regPrev.setValue(endAt);

    // store the input heap
    SymHeap input(sh);
    Trace::waiveCloneOperation(input);

    // disconnect the region from the remainder of the list
    dlsNext.setValue(VAL_NULL);

    // register pre/post pair for pop_back_by_ref() on a 2+ list
    OpFootprint *fp = new OpFootprint(input, /* output */ sh);
    fp->outArgs.push_back(reg); // disconnected region exists
    tpl->addFootprint(fp);
}

    return tpl;
}

OpTemplate* createInsert(TplFactory &fact)
{
    OpTemplate *tpl = new OpTemplate("insert_before", SHAPE_DLL_NULL);
    SymHeap sh(fact.createHeap());
    const TOffset offHead = fact.headAt();
    const TOffset offNext = fact.nextAt();
    const TOffset offPrev = fact.prevAt();

    // allocate a pair of DLSs
    const TObjId dlsPrefix = fact.createObj(&sh, OK_DLS);
    const TObjId regMiddle = fact.createObj(&sh, OK_REGION);
    const TObjId dlsSuffix = fact.createObj(&sh, OK_DLS);

    // obtain handles to prev/next pointers
    const PtrHandle dlsPrefixBack(sh, dlsPrefix, offPrev);
    const PtrHandle dlsPrefixNext(sh, dlsPrefix, offNext);
    const PtrHandle regMiddleBack(sh, regMiddle, offPrev);
    const PtrHandle regMiddleNext(sh, regMiddle, offNext);
    const PtrHandle dlsSuffixBack(sh, dlsSuffix, offPrev);
    const PtrHandle dlsSuffixNext(sh, dlsSuffix, offNext);

    // obtain addresses to the region and end/begin of DLS prefix/suffix
    const TValId dlsPrefixEnd = sh.addrOfTarget(dlsPrefix, TS_LAST,   offHead);
    const TValId regMiddleAt  = sh.addrOfTarget(regMiddle, TS_REGION, offHead);
    const TValId dlsSuffixBeg = sh.addrOfTarget(dlsSuffix, TS_FIRST,  offHead);

    // make a list of those DLSs
    dlsPrefixBack.setValue(VAL_NULL);
    dlsPrefixNext.setValue(regMiddleAt);
    regMiddleBack.setValue(dlsPrefixEnd);
    regMiddleNext.setValue(dlsSuffixBeg);
    dlsSuffixBack.setValue(regMiddleAt);
    dlsSuffixNext.setValue(VAL_NULL);

    // take the current heap snapshot as an input of the template
    SymHeap input(sh);
    Trace::waiveCloneOperation(input);

    // allocate a region to be inserted
    const TObjId regBefore = fact.createObj(&sh, OK_REGION);
    const TValId regBeforeAt = sh.addrOfTarget(regBefore, TS_REGION, offHead);
    const PtrHandle regBeforeBack(sh, regBefore, offPrev);
    const PtrHandle regBeforeNext(sh, regBefore, offNext);

    // insert the region between prefix and suffix
    dlsPrefixNext.setValue(regBeforeAt);
    regMiddleBack.setValue(regBeforeAt);
    regBeforeBack.setValue(dlsPrefixEnd);
    regBeforeNext.setValue(regMiddleAt);

    // finally define the insert_before() template!
    OpFootprint *fp = new OpFootprint(input, /* output */ sh);
    fp->inArgs.push_back(regMiddle);
    fp->outArgs.push_back(regBefore);
    tpl->addFootprint(fp);

    // special variant for inserting the first node and/or before last
    OpFootprint *fpNoPrefix = new OpFootprint(*fp);
    OpFootprint *fpNoSuffix = new OpFootprint(*fp);
    OpFootprint *fpToSingle = new OpFootprint(*fp);

    // inserting first (output)
    sh.objInvalidate(dlsPrefix);
    regBeforeBack.setValue(VAL_NULL);
    fpNoPrefix->output = sh;
    Trace::waiveCloneOperation(fpNoPrefix->output);

    // inserting first (input)
    sh.objInvalidate(regBefore);
    regMiddleBack.setValue(VAL_NULL);
    fpNoPrefix->input = sh;
    Trace::waiveCloneOperation(fpNoPrefix->input);

    sh = fp->output;
    Trace::waiveCloneOperation(sh);

    // inserting before last (output)
    sh.objInvalidate(dlsSuffix);
    PtrHandle(sh, regMiddle, offNext).setValue(VAL_NULL);
    fpNoSuffix->output = sh;
    Trace::waiveCloneOperation(fpNoSuffix->output);

    // inserting before last (input)
    sh.objInvalidate(regBefore);
    PtrHandle(sh, regMiddle, offPrev).setValue(dlsPrefixEnd);
    PtrHandle(sh, dlsPrefix, offNext).setValue(regMiddleAt);
    fpNoSuffix->input = sh;
    Trace::waiveCloneOperation(fpNoSuffix->input);

    sh = fp->output;
    Trace::waiveCloneOperation(sh);

    // inserting into singleton list (output)
    sh.objInvalidate(dlsPrefix);
    sh.objInvalidate(dlsSuffix);
    PtrHandle(sh, regBefore, offPrev).setValue(VAL_NULL);
    PtrHandle(sh, regMiddle, offNext).setValue(VAL_NULL);
    fpToSingle->output = sh;
    Trace::waiveCloneOperation(fpToSingle->output);

    // inserting into singleton list (input)
    sh.objInvalidate(regBefore);
    PtrHandle(sh, regMiddle, offPrev).setValue(VAL_NULL);
    fpToSingle->input = sh;
    Trace::waiveCloneOperation(fpToSingle->input);

    tpl->addFootprint(fpNoPrefix);
    tpl->addFootprint(fpNoSuffix);
    tpl->addFootprint(fpToSingle);
    return tpl;
}

// FIXME: createInsert() and createRemove() have a lot of identical code
OpTemplate* createRemove(TplFactory &fact)
{
    OpTemplate *tpl = new OpTemplate("remove", SHAPE_DLL_NULL);
    SymHeap sh(fact.createHeap());
    const TOffset offHead = fact.headAt();
    const TOffset offNext = fact.nextAt();
    const TOffset offPrev = fact.prevAt();

    // allocate a pair of DLSs
    const TObjId dlsPrefix = fact.createObj(&sh, OK_DLS);
    const TObjId regMiddle = fact.createObj(&sh, OK_REGION);
    const TObjId dlsSuffix = fact.createObj(&sh, OK_DLS);

    // obtain handles to prev/next pointers
    const PtrHandle dlsPrefixBack(sh, dlsPrefix, offPrev);
    const PtrHandle dlsPrefixNext(sh, dlsPrefix, offNext);
    const PtrHandle regMiddleBack(sh, regMiddle, offPrev);
    const PtrHandle regMiddleNext(sh, regMiddle, offNext);
    const PtrHandle dlsSuffixBack(sh, dlsSuffix, offPrev);
    const PtrHandle dlsSuffixNext(sh, dlsSuffix, offNext);

    // obtain addresses to the region and end/begin of DLS prefix/suffix
    const TValId dlsPrefixEnd = sh.addrOfTarget(dlsPrefix, TS_LAST,   offHead);
    const TValId regMiddleAt  = sh.addrOfTarget(regMiddle, TS_REGION, offHead);
    const TValId dlsSuffixBeg = sh.addrOfTarget(dlsSuffix, TS_FIRST,  offHead);

    // make a list of those DLSs
    dlsPrefixBack.setValue(VAL_NULL);
    dlsPrefixNext.setValue(regMiddleAt);
    regMiddleBack.setValue(dlsPrefixEnd);
    regMiddleNext.setValue(dlsSuffixBeg);
    dlsSuffixBack.setValue(regMiddleAt);
    dlsSuffixNext.setValue(VAL_NULL);

    // take the current heap snapshot as an input of the template
    SymHeap input(sh);
    Trace::waiveCloneOperation(input);

    // remove regMiddle
    sh.objInvalidate(regMiddle);
    dlsPrefixNext.setValue(dlsSuffixBeg);
    dlsSuffixBack.setValue(dlsPrefixEnd);

    // finally define the remove() template!
    OpFootprint *fp = new OpFootprint(input, /* output */ sh);
    fp->inArgs.push_back(regMiddle);
    tpl->addFootprint(fp);

    // special cases
    OpFootprint *fpNoPrefix = new OpFootprint(*fp);
    OpFootprint *fpNoSuffix = new OpFootprint(*fp);
    OpFootprint *fpOfSingle = new OpFootprint(*fp);

    // removing first (input)
    sh = fp->input;
    Trace::waiveCloneOperation(sh);
    sh.objInvalidate(dlsPrefix);
    regMiddleBack.setValue(VAL_NULL);
    fpNoPrefix->input = sh;
    Trace::waiveCloneOperation(fpNoPrefix->input);

    // removing first (output)
    sh.objInvalidate(regMiddle);
    dlsSuffixBack.setValue(VAL_NULL);
    fpNoPrefix->output = sh;
    Trace::waiveCloneOperation(fpNoPrefix->output);

    // removing last (input)
    sh = fp->input;
    Trace::waiveCloneOperation(sh);
    sh.objInvalidate(dlsSuffix);
    regMiddleNext.setValue(VAL_NULL);
    fpNoSuffix->input = sh;
    Trace::waiveCloneOperation(fpNoSuffix->input);

    // removing last (output)
    sh.objInvalidate(regMiddle);
    dlsPrefixNext.setValue(VAL_NULL);
    fpNoSuffix->output = sh;
    Trace::waiveCloneOperation(fpNoSuffix->output);

    // removing singleton (input)
    sh = fp->input;
    Trace::waiveCloneOperation(sh);
    sh.objInvalidate(dlsPrefix);
    sh.objInvalidate(dlsSuffix);
    regMiddleBack.setValue(VAL_NULL);
    regMiddleNext.setValue(VAL_NULL);
    fpOfSingle->input = sh;
    Trace::waiveCloneOperation(fpOfSingle->input);

    // removing last (output)
    sh.objInvalidate(regMiddle);
    fpOfSingle->output = sh;
    Trace::waiveCloneOperation(fpOfSingle->output);

    tpl->addFootprint(fpNoPrefix);
    tpl->addFootprint(fpNoSuffix);
    tpl->addFootprint(fpOfSingle);
    return tpl;
}

OpTemplate* createClear2(TplFactory &fact)
{
    OpTemplate *tpl = new OpTemplate("clear2", SHAPE_DLL_NULL);

    // create an empty heap as the output
    SymHeap sh(fact.createHeap());
    SymHeap output(sh);
    Trace::waiveCloneOperation(output);

    // allocate a pair of regions
    const TObjId reg1 = fact.createObj(&sh, OK_REGION);
    const TObjId reg2 = fact.createObj(&sh, OK_REGION);

    // acquire handles to their prev/next fields
    const PtrHandle reg1Prev(sh, reg1, fact.prevAt());
    const PtrHandle reg1Next(sh, reg1, fact.nextAt());
    const PtrHandle reg2Prev(sh, reg2, fact.prevAt());
    const PtrHandle reg2Next(sh, reg2, fact.nextAt());

    // connect them as a list
    reg1Prev.setValue(VAL_NULL);
    reg1Next.setValue(sh.addrOfTarget(reg2, TS_REGION, fact.headAt()));
    reg2Prev.setValue(sh.addrOfTarget(reg1, TS_REGION, fact.headAt()));
    reg2Next.setValue(VAL_NULL);

    // register pre/post pair for clear2()
    tpl->addFootprint(new OpFootprint(/* input */ sh, output));
    return tpl;
}

bool loadDefaultOperations(OpCollection *pDst, const CodeStorage::Storage &stor)
{
    if (stor.types.dataPtrSizeof() <= 0) {
        CL_BREAK_IF("loadDefaultOperations() failed to resolve sizeof(void *)");
        return false;
    }

    TplFactory fact(stor);
    for (int i = LS_BACK; i >= LS_FRONT; --i) { // reverse search order
        const EListSide side = static_cast<EListSide>(i);
#if 0
        pDst->addTemplate(createPushByRef(fact, side));
        pDst->addTemplate(createPushByVal(fact, side));
        pDst->addTemplate(createPop(fact, side));
        pDst->addTemplate(createPopRef(fact, side));
#endif

#if 1
        pDst->addTemplate(createCDLLPushByVal(fact, side));
        pDst->addTemplate(createCDLLPop(fact, side));
#endif

    }

#if 0
    pDst->addTemplate(createClear2(fact));
    pDst->addTemplate(createInsert(fact));
    pDst->addTemplate(createRemove(fact));
#endif

    return true;
}

} // namespace AdtOp
