/*
 * Copyright (C) 2013 Kamil Dudka <kdudka@redhat.com>
 *
 * This file is part of predator.
 *
 * predator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * predator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with predator.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"
#include "fixed_point.hh"
#include "adt_op_match.hh"

#include "adt_op_meta.hh"
#include "cont_shape_seq.hh"
#include "symseg.hh"                // for objMinLength()

#include <cl/cl_msg.hh>
#include <cl/storage.hh>            // for CodeStorage::TypeDb::dataPtrSizeof()

#include <algorithm>                // for std::reverse

#include <boost/foreach.hpp>

namespace AdtOp {

bool debuggingTplMatch = true;

#if 1
#define TM_DEBUG(msg) do {              \
    if (!AdtOp::debuggingTplMatch)      \
        break;                          \
    CL_DEBUG(msg);                      \
} while (0)
#else
#define TM_DEBUG(msg) do {              \
    if (!AdtOp::debuggingTplMatch)      \
        break;                          \
    std::cout << "\x1b[1mDEBUGt:\x1b[0m " << msg << std::endl;                      \
} while (0)
#endif

#if 1
typedef FixedPoint::TObjectMapper                   TObjectMapper;
typedef FixedPoint::TLocIdx                         TLocIdx;
typedef FixedPoint::THeapIdx                        THeapIdx;
typedef FixedPoint::THeapIdent                      THeapIdent;
typedef FixedPoint::TShapeIdent                     TShapeIdent;

//FIXME: duplicated output operators
/// output
inline std::ostream & operator << (std::ostream &s, THeapIdent hi) {
    return s << "Heap(" << hi.first << "/" << hi.second << ")";
}

/// output
inline std::ostream & operator << (std::ostream &s, TShapeIdent si) {
    return s << "Shape(" << si.first << ", " << si.second << ")";
}

#else
// no Koenig lookup? for op<<
using FixedPoint::TObjectMapper;
using FixedPoint::TLocIdx      ;
using FixedPoint::THeapIdx     ;
using FixedPoint::THeapIdent   ;
using FixedPoint::TShapeIdent  ;
using namespace FixedPoint;
#endif

struct MatchCtx {
    TMatchList                     &matchList;
    const OpCollection             &opCollection;
    const TProgState               &progState;
    FixedPoint::TShapeSeqList       shapeSeqs;

    MatchCtx(
            TMatchList             &matchList_,
            const OpCollection     &opCollection_,
            const TProgState       &progState_):
        matchList(matchList_),
        opCollection(opCollection_),
        progState(progState_)
    {
        FixedPoint::collectShapeSequences(&shapeSeqs, progState);
    }
};

unsigned countObjects(const SymHeap &sh)
{
    TObjList objs;
    sh.gatherObjects(objs);
    return objs.size();
}

enum EContext {
    C_TEMPLATE,
    C_PROGRAM,
    C_TOTAL
};

/// map and remove corresponding regions from end of lists
void processRegSuffix(
        TObjectMapper              *pMap,
        TObjList                    pObjLists[1][C_TOTAL],
        const SymHeap              &shTpl,
        const SymHeap              &shProg)
{
    TObjList &list0 = (*pObjLists)[0]; // template objects
    TObjList &list1 = (*pObjLists)[1]; // program objects

    while (!list0.empty() && !list1.empty()) {
        const TObjId obj0 = list0.back();
        const TObjId obj1 = list1.back();

        // check that both objects are regions
        if (OK_REGION !=  shTpl.objKind(obj0))
            break;
        if (OK_REGION != shProg.objKind(obj1))
            break;

        // remember the mapping and remove the pair of regions
        pMap->insert(obj0, obj1);
        list0.pop_back();
        list1.pop_back();
    }
}

// TODO: clean-up interface, rename to reverseObjLists
void swapObjLists(TObjList pObjLists[1][C_TOTAL])
{
    for (unsigned u = 0U; u < C_TOTAL; ++u) {
        TObjList &objList = (*pObjLists)[u];
        std::reverse(objList.begin(), objList.end());
    }
}

typedef std::vector<TObjectMapper>                  TObjMapList;
typedef TObjList                                    TObjListArray[C_TOTAL];

/// create mapping tplObj->progObj based on Obj positions in objLists
void insertToObjMap(
        TObjectMapper              *pObjMap,
        const TObjListArray        &objLists,
        const int                   tplObjIdx,
        const int                   progObjIdx)
{
    const TObjId objTpl = objLists[C_TEMPLATE].at(tplObjIdx);
    const TObjId objProg = objLists[C_PROGRAM].at(progObjIdx);
    pObjMap->insert(objTpl, objProg);
}

/// find all? possible mappings of template objects to program objects
bool resolveObjectMapping(
        TObjMapList                *pDst,
        const TObjectMapper        &objMapOrig,
        const SymHeap              &shProg,
        const SymHeap              &shTpl,
        const TObjListArray        &objLists)
{
    const int tplObjCnt = objLists[C_TEMPLATE].size();
    const int progObjCnt = objLists[C_PROGRAM].size();

//# std::cout << "[function] resolveObjectMapping: " << " tplObjCnt=" << tplObjCnt << ", progObjCnt=" << progObjCnt << std::endl;

    CL_BREAK_IF(!tplObjCnt);
    if (tplObjCnt == 1) {
        const TObjId objTpl = objLists[C_TEMPLATE].front();
        const EObjKind kind = shTpl.objKind(objTpl);
        if (kind == OK_DLS || progObjCnt == 1) {
            // map objTpl (the only tpl object) to the remaining shProg objects
            TObjectMapper objMap(objMapOrig);
            BOOST_FOREACH(const TObjId objProg, objLists[C_PROGRAM])
                objMap.insert(objTpl, objProg);
            pDst->push_back(objMap);
//# std::cout << "r          resolveObjectMapping: " << (pDst->size()>0?"OK ":"Fail ") << pDst->size() << std::endl;
            return true;
        }
    }

//std::cout << "i  objLists[C_TEMPLATE]=" << objLists[C_TEMPLATE] << std::endl;
//std::cout << "i  objLists[C_PROGRAM]=" << objLists[C_PROGRAM] << std::endl;


    bool dlsAtBeg = false;
    bool dlsAtEnd = false;
    for (int i = 0; i < tplObjCnt; ++i) {
        const TObjId objTpl = objLists[C_TEMPLATE][i];
        if (shTpl.objKind(objTpl) != OK_DLS)
            continue;

        if (i==0)
            dlsAtBeg = true;
        else if (i + 1 == tplObjCnt)
            dlsAtEnd = true;
        else {
            // template DLS in the middle not supported for now
            // =Error in template
            CL_BREAK_IF("template problem: DLS in the middle not supported for now");
            return false;
        }
    }

// TODO: PROBLEM: T[dls,r] P[r,dls,dls] = NO MATCH
// TODO: Possible solution: if R-D and R-D-R-D..., we can match all remaining objects to D
    for (int i = 1; i < progObjCnt - 1; ++i) {
        const TObjId objProg = objLists[C_PROGRAM][i];
        if (shProg.objKind(objProg) == OK_DLS) {
            // program DLS in the middle not supported for now
//# std::cout << "ERR ---------- program DLS in the middle not supported for now" << std::endl;
            return false;
        }
    }

    const int cntDiff = progObjCnt - tplObjCnt;
    if (0==cntDiff && !dlsAtBeg && !dlsAtEnd) {
        // nothing to merge here
//# std::cout << "ERR ----- nothing to merge here" << std::endl;
        return false;
    }

    CL_BREAK_IF(cntDiff < 0); // (progObjCnt < tplObjCnt) -> error

    const int mergeLeftMax = (dlsAtBeg) ? cntDiff : 0;
    const int mergeLeftMin = (dlsAtEnd) ? 0 : cntDiff;

    for (int mergeLeft = mergeLeftMin; mergeLeft <= mergeLeftMax; ++mergeLeft) {
        TObjectMapper objMap(objMapOrig);
        int idxTpl = 0;
        int idxProg = 0;

        // merge DLS in the left with the prefix
        for (; idxProg <= mergeLeft; ++idxProg)
            insertToObjMap(&objMap, objLists, idxTpl, idxProg);

        // map the inner regions 1:1
        for (++idxTpl; idxTpl < (tplObjCnt - 1); ++idxTpl, ++idxProg)
            insertToObjMap(&objMap, objLists, idxTpl, idxProg);

        // merge DLS in the right with the prefix
        for (; idxProg < progObjCnt; ++idxProg)
            insertToObjMap(&objMap, objLists, idxTpl, idxProg);

        pDst->push_back(objMap);
    }

//# std::cout << "r          resolveObjectMapping: " << (pDst->size()>0?"OK ":"Fail ") << pDst->size() << std::endl;

    return !pDst->empty();
}

/// search for anchor heap in "red line"
bool matchAnchorHeapCore(
        TObjMapList                *pDst,
        const SymHeap              &shProg,
        const SymHeap              &shTpl,
        const Shape                &csProg,
        const Shape                &csTpl)
{
//std::cout << "[function] matchAnchorHeapCore: " << std::endl;
//std::cout << "|p| "; csProg.dump(std::cout,shProg);
//std::cout << "|t| "; csTpl.dump(std::cout,shTpl);

    if (csTpl.length != countObjects(shTpl)) {
//# std::cout << "!          matchAnchorHeapCore: (csTpl.length!=countObjects(shTpl))==(" << csTpl.length << "!=" << countObjects(shTpl) << ")==FALSE  unsupported anchor heap in a template" << std::endl;
        CL_BREAK_IF("unsupported anchor heap in a template");
        return false;
    }

    // resolve list of objects belonging to containers shapes
    TObjList objLists[C_TOTAL];
    objListByShape(&objLists[C_TEMPLATE], shTpl, csTpl);        // objects in template shape
    objListByShape(&objLists[C_PROGRAM], shProg, csProg);       //            program
    CL_BREAK_IF(objLists[C_TEMPLATE].empty());

// TODO: create function stripEndRegions, remove both next functions
    // handle matching regions at both end-points
    TObjectMapper objMap;
    for (unsigned u = 0; u < 2; ++u) {
        swapObjLists(&objLists);        // reverse both lists
        processRegSuffix(&objMap, &objLists, shTpl, shProg);
    }

    if (objLists[C_TEMPLATE].empty() && objLists[C_PROGRAM].empty()) {
        // all regions were mapped and nothing remains
        pDst->push_back(objMap);
        return true;
    }

    const unsigned tplObjectCount = objLists[C_TEMPLATE].size();
    if (tplObjectCount==0 || objLists[C_PROGRAM].size() < tplObjectCount) {
        // we are no longer able to map the remaining objects
//# std::cout << "-          matchAnchorHeapCore: object count mismatch" << std::endl;
        return false;
    }

    // compute total minimal length of the remaining template objects
    TMinLen lenTpl = 0;
    BOOST_FOREACH(const TObjId objTpl, objLists[C_TEMPLATE])
        lenTpl += objMinLength(shTpl, objTpl);

    // compute total minimal length of the remaining program objects
    TMinLen lenProg = 0;
    BOOST_FOREACH(const TObjId objProg, objLists[C_PROGRAM])
        lenProg += objMinLength(shProg, objProg);

    if (lenProg < lenTpl) {
        // the program configuration does not guarantee sufficient list length
//# std::cout << "-          matchAnchorHeapCore: lenProg<lenTpl" << std::endl;
        return false;
    }

    resolveObjectMapping(pDst, objMap, shProg, shTpl, objLists);
    return !pDst->empty();
}

// search for anchor heap in "red-line"
bool matchAnchorHeap(
        TMatchList                 *pMatchList,
        MatchCtx                   &ctx,
        const OpTemplate           &tpl,
        const OpFootprint          &fp,
        const TFootprintIdent      &fpIdent,
        const TShapeIdent          &shIdent)
{
    // resolve program state and shape
    using namespace FixedPoint;
    THeapIdent heapId = shIdent.first;
    const SymHeap &shProg = *heapByIdent(ctx.progState, shIdent.first);
    const Shape &csProg = *shapeByIdent(ctx.progState, shIdent);

//# std::cout << "[function] matchAnchorHeap: input " << heapId << std::endl;

    // check search direction
    bool reverse = false;
    const ESearchDirection sd = tpl.searchDirection();
    switch (sd) {
        case SD_FORWARD:
            break;

        case SD_BACKWARD:
            reverse = true;
            break;

        default:
            CL_BREAK_IF("matchAnchorHeap() got invalid search direction");
    }

    // resolve template state and shape list
    const SymHeap &shTpl = (reverse)
        ? fp.output
        : fp.input;
    const TShapeListByHeapIdx &csTplListByIdx = (reverse)
        ? tpl.outShapes()
        : tpl.inShapes();

    // check the count of container shapes in the template
    const TShapeList &csTplList = csTplListByIdx[fpIdent./* footprint */second];
    if (1U != csTplList.size()) {
        CL_BREAK_IF("unsupported count of shapes in matchAnchorHeap()");
        return false;
    }

    // resolve objMap by search direction
    const EFootprintPort port = (reverse)
        ? FP_DST
        : FP_SRC;

    // perform an object-wise match
    const Shape &csTpl = csTplList.front();
    TObjMapList objMapList;
    if (!matchAnchorHeapCore(&objMapList, shProg, shTpl, csProg, csTpl)) {
//# std::cout << "-          matchAnchorHeap: FALSE" << std::endl;
        return false;
    }

    CL_BREAK_IF(objMapList.empty());

    // successful match!
    FootprintMatch fmProto(fpIdent);
    fmProto.props = csProg.props;
    fmProto.tplProps = csTpl.props;
    fmProto.matchedHeaps.push_back(shIdent.first);

//# std::cout << std::endl;
//# std::cout << "|p| "; csProg.dump(std::cout,shProg);
//# std::cout << "|t| "; csTpl.dump(std::cout,shTpl);
//# std::cout << "    objMapList.size=" << objMapList.size() << ", contents= " << std::endl;

    BOOST_FOREACH(const TObjectMapper &objMap, objMapList) {
//# std::cout << "|m| "; sl_dump(objMap);        // ======================================LOG
        FootprintMatch fm(fmProto);
        fm.objMap[port] = objMap;
        pMatchList->push_back(fm);
    }

//# std::cout << std::endl;
//# std::cout << "+          matchAnchorHeap: TRUE " << heapId << " for " << tpl.name() << "[" << fpIdent.second << "]"<< std::endl;

    return true;
}

// select first object in objList (size>1, unsorted sequence) of source objects
// based on direction ts from which we acces the sequence
// TODO: check for cDLL functionality
TObjId selectMappedObjByTs(
        const SymHeap              &sh,
        const BindingOff           &bOff,
        const TObjList             &objList,
        const ETargetSpecifier      ts)
{
//# std::cout << "[function] selectMappedObjByTs: objList=" << objList << " ts=" << ts << std::endl;

    switch (ts) {
        case TS_FIRST:
        case TS_LAST:
            break;

        default:
            TM_DEBUG("selectMappedObjByTs() got invalid target specifier");
            return OBJ_INVALID;
    }

    // start with all objects as candidates for fst/lst
    TObjSet allObjs, cObjs;
    BOOST_FOREACH(const TObjId obj, objList) {
        allObjs.insert(obj);
        cObjs.insert(obj);
    }

    // prepare offsets for direction ts
    const TOffset offNext = (TS_FIRST == ts) ? bOff.next : bOff.prev;
    const TOffset offPrev = (TS_LAST  == ts) ? bOff.next : bOff.prev;

    SymHeap &shWritable = const_cast<SymHeap &>(sh);

// TODO: use some function shared with shape discovery code
// BUG: does not work correctly - it does not check pairs
//      example: V-type shape, links: \=next /=prev
// BUG for cDLL - there can be 2 objects => SYMMETRIC
//     (removes both objects: objNext and objPrev)
    BOOST_FOREACH(const TObjId obj, objList) {
        const TObjId objNext = nextObj(shWritable, obj, offNext);
        cObjs.erase(objNext);   // remove next object iff in set

        const TObjId objPrev = nextObj(shWritable, obj, offPrev);
        if (hasKey(allObjs, objPrev))
            cObjs.erase(obj);   // remove obj if predecessor in allObjs
    }

// HACK for cDLL of length 2 -- result of concretization
    if(cObjs.size()==0 && objList.size()==2) {
        // "FIRST" object does not exist, maybe cDLS concretization
        // D-R or R-D
        int i = 0;
        BOOST_FOREACH(const TObjId obj, objList) {
            // get opposite object
            const TObjId obj1 = (i++==0)? objList[1] : objList[0];

            const TObjId objNext = nextObj(shWritable, obj, offNext);
            const TObjId objNextPrev = nextObj(shWritable, objNext, offPrev);
            if(objNext==obj1 && objNextPrev==obj) {
                 // obj<=>obj1 doubly-linked in the requested direction
                 // return first object in the direction

// INFO: we do not know which object is first - may be loop for cDLL of size 2
//       we use concretization info from symbolic execution
// TODO: check
                if(ts==sh.getLastConcretizedRegDir(obj)) {      // HACK2
                    // is region AND concretized on the requested side
                    cObjs.insert(obj);
                } else if (OK_DLS==sh.objKind(obj)){
                    // is DLS on the requested side
                    cObjs.insert(obj);
                }
            }
        } // for
//# std::cout << "Workaround: map to 2 objects of cDLL, result.size=" << cObjs.size() <<std::endl;
    }

    // INFO: all srcobjects checked for prev/next linkage
    //       first object remains in cObjs iff single sequence

    switch (cObjs.size()) {
        case 0U:        // not complete sequence
            // we lost trace of the object we have been looking for
//# std::cout << "ERROR: we lost trace in the input object sequence" <<std::endl;
//abort();

            return OBJ_INVALID;

        case 1U:        // single sequence found
            return *cObjs.begin();      // single object in set

        default:        // more sequences found
            TM_DEBUG("unsupported ID mapping in selectMappedObjByTs()");
//abort();
            return OBJ_INVALID;
    }
}

/// deduce dls-target-spec for region using shape props
ETargetSpecifier tsByOffset(const TOffset off, const ShapeProps &props)
{
    const BindingOff &bf = props.bOff;
    if (off == bf.next)
        return TS_LAST;

    if (off == bf.prev)
        return TS_FIRST;

    CL_BREAK_IF("tsByOffset() failed to resolve target specifier");
    return TS_INVALID;
}

/// change object *pObj to start object of mapped subDLS with props and direction ts
bool relocSingleObj(
        TObjId                     *pObj,
        const ETargetSpecifier      ts,
        const TObjectMapper        &objMap,
        const ShapeProps           &props,
        const SymHeap              &sh,
        const bool target)
{
//# std::cout << "[function] relocSingleObj: obj1=" << *pObj << " ts=" << ts << (target?" TARGET":" OBJ") << std::endl;

    // do not relocate ID of special objects
    TObjId obj1 = *pObj;
    switch (obj1) {
        case OBJ_INVALID:
            CL_BREAK_IF("relocSingleObj() got OBJ_INVALID");
            return false;

        case OBJ_NULL:
            return true;

        default:
            break;
    }

    // query the object map to obtain the resulting object ID
    TObjList objList;
    objMap.query<D_LEFT_TO_RIGHT>(&objList, obj1);

    TObjId obj = OBJ_INVALID;
    if(objList.size()==1U) {
        obj = objList.front();      // unambiguous ID mapping
    } else {
        obj = selectMappedObjByTs(sh, props.bOff, objList, ts);
    }

    if (OBJ_INVALID == obj) {
//# std::cout << "-          relocSingleObj: FALSE" << std::endl;
        return false;
    }

    *pObj = obj;
//# std::cout << "+          relocSingleObj: TRUE, obj2=" << obj << std::endl;
    return true;
}

/// relocate obj and tgtObj in each MetaOperation (MOP1->MOP2)
bool relocObjsInMetaOps(
        TMetaOpSet                 *pMetaOps,
        const TObjectMapper        &objMap,
        const ShapeProps           &props,
        const SymHeap              &sh)
{
//# std::cout << "[function] relocObjsInMetaOps: " << std::endl;
//# std::cout << "RELOC/INMAP=" << *pMetaOps << std::endl;

    const TMetaOpSet &src = *pMetaOps;
    TMetaOpSet dst;

    BOOST_FOREACH(MetaOperation mo, src) {
        MetaOperation mo_orig = mo;

        ETargetSpecifier ts = TS_INVALID;
        if (MO_SET == mo.code) {        // MOP has target part
            ts = tsByOffset(mo.off, props);
            const ETargetSpecifier tgtTs = (TS_REGION == mo.tgtTs)
                ? ts
                : mo.tgtTs;
            // INFO: maps mo.tgtObj, if mapped to subDLL, take correct end
            if (!relocSingleObj(&mo.tgtObj, tgtTs, objMap, props, sh, true)) {
//# std::cout << "-          relocObjsInMetaOps: returns FALSE" << std::endl;
                return false;
            }
        } // if MO_SET

        // INFO: maps mo.obj, if mapped to subDLL, take correct end
        if (!relocSingleObj(&mo.obj, ts, objMap, props, sh, false)) {
//# std::cout << "-          relocObjsInMetaOps: returns FALSE" << std::endl;
            return false;
        }

#if 1
// check if changed
if(mo!=mo_orig) {
//# std::cout << "CHANGE     relocObjsInMetaOps: ORIG" << mo_orig << std::endl;
if(mo.tgtTs != mo_orig.tgtTs) {
//# std::cout << "           tgtTs CHANGED: " << mo_orig.tgtTs << "->" << mo.tgtTs << std::endl;
}
// print what is added
//# std::cout << "+          relocObjsInMetaOps: ADD " << mo << std::endl;
}
#endif

        dst.insert(mo);
    } // end MetaOperation loop

    pMetaOps->swap(dst);
//# std::cout << "+++        relocObjsInMetaOps: returns TRUE" << std::endl;
//# std::cout << "RELOC/OUTMAP=" << *pMetaOps << std::endl;
    return true;
}

/// change MOP next/prev offsets from template props to program props
void relocFieldOffset(MetaOperation *pMetaOp, const FootprintMatch &fm)
{
    const BindingOff &bfTpl = fm.tplProps.bOff;
    const BindingOff &bfProg = fm.props.bOff;
    TOffset &off = pMetaOp->off;

    CL_BREAK_IF(bfTpl.next == bfTpl.prev);

    if (off == bfTpl.next)
        off = bfProg.next;
    else if (off == bfTpl.prev)
        off = bfProg.prev;

    if (MO_SET != pMetaOp->code || OBJ_NULL == pMetaOp->tgtObj)
        return;

    // INFO: MOP has target part (non NULL)
    // TODO: check this!        it expects bfTpl.head==0 ?
    if (bfProg.head!=0 && (bfTpl.head == pMetaOp->tgtOff))
        // relocate target offset according to head offset
        pMetaOp->tgtOff = bfProg.head;
}

/// relocate MOPs offsets according to program
void relocOffsetsInMetaOps(TMetaOpSet *pMetaOps, const FootprintMatch &fm)
{
    const TMetaOpSet &src = *pMetaOps;
    TMetaOpSet dst;

    BOOST_FOREACH(MetaOperation mo, src) {
        switch (mo.code) {
            case MO_SET:
            case MO_UNSET:
                relocFieldOffset(&mo, fm);
                break;

            default:
                break;
        }

        dst.insert(mo);         // maybe modified MOP
    }

    pMetaOps->swap(dst);        // return translated MOP set
}

typedef const FixedPoint::TraceEdge                *TEdgePtr;

typedef std::vector<THeapIdent>                     THeapIdentList;
typedef std::vector<TObjectMapper>                  TObjectMapperList;

/// create list of heaps and associated maps in correct order
void collectNextHeaps(
        THeapIdentList             *pHeapList,
        TObjectMapperList          *pObjMapList,
        const THeapIdent            heapCurrent,
        const TProgState           &progState,
        const ESearchDirection      sd)
{
    using namespace FixedPoint;
    const LocalState &locState = progState[heapCurrent./* loc */first];

    // check search direction
    bool reverse = false;
    switch (sd) {
        case SD_FORWARD:
            break;

        case SD_BACKWARD:
            reverse = true;
            break;

        default:
            CL_BREAK_IF("jumpToNextHeap() got invalid search direction");
    }

    const TEdgeListByHeapIdx &eListByHeapIdx = (reverse)
        ? locState.traceInEdges
        : locState.traceOutEdges;

    const TTraceEdgeList &edgeList = eListByHeapIdx[heapCurrent./* sh */second];
    BOOST_FOREACH(const TEdgePtr te, edgeList) {
        const THeapIdent heapNext = (reverse)
            ? te->src
            : te->dst;

        TObjectMapper objMap = te->objMap;
        if (reverse)
            objMap.flip();

        pHeapList->push_back(heapNext);
        pObjMapList->push_back(objMap);
    }
}

bool removeOpFrom(TMetaOpSet *pLookup, const SymHeap &sh0, MetaOperation mo)
{
//# std::cout << "[function] removeOpFrom: mops#=" << pLookup->size() << " mo=" << std::endl;
//# std::cout << "  " << mo << " to match in MOPset of size=" << pLookup->size() << std::endl;

// TODO: move next check to some checking function
    switch (mo.code) {
        case MO_ALLOC:
        case MO_FREE:
            if (OK_REGION != sh0.objKind(mo.obj)) {
                // only regions can be allocated/freed in a single step
//# std::cout << "-          removeOpFrom: FALSE  only regions can be allocated/freed in a single step" << std::endl;
                return false;
            }
        default:
            break;
    }

/////////////////////////////////////////////////////////////////////////////
// PRINT metaops got from template
//# std::cout << "X ----------------------- MATCH? -----------------------------" << std::endl;
//# std::cout << *pLookup;
//# std::cout << "X ------------------------------------------------------------" << std::endl;
/////////////////////////////////////////////////////////////////////////////

    if (1U == pLookup->erase(mo))
        return true;            // program micro-operation is equal to some template MOP

    if (MO_SET != mo.code) {
//# std::cout << "-          removeOpFrom: FALSE  (MO_SET != mo.code) - not pointer assignment" << std::endl;
        return false;           // not pointer assignment
    }

    const EObjKind kind = sh0.objKind(mo.tgtObj);
    if (OK_REGION != kind) {
#if 0
//# std::cout << "-          removeOpFrom: FALSE  (OK_REGION != kind)" << std::endl;
        return false;         // in template target is NOT REGION
#else
///////////////////////////////////////////////////////////////////////////
/// WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
///////////////////////////////////////////////////////////////////////////
/// HACK? I am NOT SURE, but pointing to ANY end of DLS is possible using
/// prev/next, further micro-operation will concretize and THEN we see...
//
///    +---------+
///    +-DLS-DLS-+
///        ^            head abstracted to first DLS end
///    +---------+
///    +-DLS-DLS-+
///          ^          head abstracted to second DLS begin
//
///  we point to head (no write, so not concretized yet)
//
/// FIXME: should work for CYCLIC DLL, TODO: NULL-DLL
///////////////////////////////////////////////////////////////////////////
/// WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
///////////////////////////////////////////////////////////////////////////
//# std::cout << "i          removeOpFrom: PROBLEM: target (OK_REGION != kind)" << std::endl;

        // fall through to REGION variant of code
#endif
    }

    // INFO:  TEMPLATE: any->R || any->DLS,   PROGRAM: any->any
    // WARNING: (maybe) problem for NULL-terminated-DLS
    // INFO: also try TS_{FIRST,LAST} for target region when template uses DLS
    ///////////////////////////////////////////////////////////////////////////

    // translate metaop's TS_REGION to TS_{FIRST,LAST} if appropriate
    mo.tgtTs = TS_FIRST;
//# std::cout << "?          removeOpFrom: TRY mo.tgtTs=" << mo.tgtTs << std::endl;
    if (1U == pLookup->erase(mo))
        return true;
    mo.tgtTs = TS_LAST;
//# std::cout << "?          removeOpFrom: TRY mo.tgtTs=" << mo.tgtTs << std::endl;
    if (1U == pLookup->erase(mo))
        return true;

//# std::cout << "---        removeOpFrom: return FALSE - metaop does not match" << std::endl;
    return false;
}

bool isIndependentOp(
        FootprintMatch             *pMatch,
        const SymHeap              &sh,
        const MetaOperation        &mo)
{
    if (MO_SET != mo.code)      // MetaOp: write pointer at offset
        return false;

    // FIXME: we assume that all fields in MetaOperation have sizeof(void *)
    const TStorRef stor = sh.stor();
    const TSizeOf psize = stor.types.dataPtrSizeof();
    const TOffset winLo = mo.off;
    const TOffset winHi = winLo + psize;
//std::cout << "=          isIndependentOp: winLo=" << winLo << " winHi=" << winHi << std::endl;

    const BindingOff &bOff = pMatch->props.bOff;
    std::set<TOffset> offs;
    offs.insert(bOff.next);
    offs.insert(bOff.prev);
    BOOST_FOREACH(const TOffset lo, offs) {
        const TOffset hi = lo + psize;

//std::cout << "=          isIndependentOp: lo=" << lo << " hi=" << hi << std::endl;
        if (lo >= winHi || hi <= winLo)         // no intersection
            continue;

        // field clash detected!
//# std::cout << "!        isIndependentOp: FALSE = field clash detected!" << std::endl;
        return false;
    }

    // proven independent
    return true;
}

/// find MOP which matches template MOPs (heap0->MOP->heap1)
bool processDiffOf(
        FootprintMatch             *pMatch,
        TMetaOpSet                 *pLookup,
        const TProgState           &progState,
        const THeapIdent            heap0,
        const THeapIdent            heap1,
        const ESearchDirection      sd)
{
//# std::cout << "\n[function] processDiffOf: " << heap0 << " --> " << heap1;
//# std::cout << " ================================ " << std::endl;

    const SymHeap &sh0 = *heapByIdent(progState, heap0);
    const SymHeap &sh1 = *heapByIdent(progState, heap1);

    // compute the difference of the pair of heaps
    TMetaOpSet metaOpsNow;
    if (!diffHeaps(&metaOpsNow, sh0, sh1, heap0.first)) {
        TM_DEBUG(" FALSE: \x1b[1m diffHeaps() has failed \x1b[0m");
        return false;
    }

    TObjSet freshObjs;
    BOOST_FOREACH(const MetaOperation &mo, metaOpsNow)
        if (MO_ALLOC == mo.code)
            freshObjs.insert(mo.obj);

    bool found = false;

    BOOST_FOREACH(const MetaOperation &mo, metaOpsNow) {

//# std::cout << "*          processDiffOf: MO-check " << mo << std::endl;

        const SymHeap &sh = (hasKey(freshObjs, mo.obj)) ? sh1 : sh0;
        const EStorageClass code = sh.objStorClass(mo.obj);
        if (isProgramVar(code))
            // we are not interested in changing non-heap variables
            continue;

        if (removeOpFrom(pLookup, sh0, mo)) {
            // INFO: successfull match, mo is implementing op

            // TODO: move to function addToMatchedHeaps(dst,h0,h1)
//# std::cout << "*          processDiffOf: ADD MATCH " << heap0 << "->" << heap1 << std::endl;
            THeapIdentSeq *pDst = &pMatch->matchedHeaps;

            if (SD_FORWARD == sd) {
                if (!pDst->empty())
                    pDst->pop_back();

                pDst->push_back(heap0);
                pDst->push_back(heap1);
            }
            else /* SD_BACKWARD */ {
                if (pDst->empty())
                    pDst->push_back(heap1);

                pDst->push_front(heap0);
            }

            found = true;
            continue;
        }

        if (isIndependentOp(pMatch, sh0, mo))
            continue;

        TM_DEBUG("failed to prove independency of a meta-operation"
                ", src=" << heap0.first << "/" << heap0.second <<
                ", dst=" << heap1.first << "/" << heap1.second);
        return false;
    }

    if (1U < pMatch->matchedHeaps.size() && !found)
        // mark the current instruction as (successfully) skipped
        pMatch->skippedHeaps.push_back(heap0);

//# std::cout << "+++        processDiffOf: TRUE";
//# std::cout << " ================================ \n" << std::endl;

    return true;
}

struct SeekContext {
    FootprintMatch                  fm;
    TMetaOpSet                      metaOpsToLookFor;
    TObjectMapper                   objMapFromTpl;
    THeapIdent                      heapCurrent;
    std::set<THeapIdent>            seen;

    SeekContext(
            const FootprintMatch   &fm_,
            const TMetaOpSet       &metaOpsToLookFor_,
            const ESearchDirection  sd):
        fm(fm_),
        metaOpsToLookFor(metaOpsToLookFor_),
        objMapFromTpl(fm_.objMap[(SD_FORWARD == sd) ? FP_SRC : FP_DST]),
        heapCurrent(fm_.matchedHeaps.front())
    {
        // pick the starting position and clear the list of matched heaps
        CL_BREAK_IF(1U != fm.matchedHeaps.size());
        seen.insert(heapCurrent);
        fm.matchedHeaps.clear();
    }
};

typedef std::set<TShapeIdent>                       TShapeIdentSet;

void insertInitShape(
        TShapeIdentSet             *pInitShapes,
        const OpTemplate           &tpl,
        const FootprintMatch       &fm)
{
    const ESearchDirection sd = tpl.searchDirection();
#ifndef NDEBUG
    const TShapeListByHeapIdx &shapeListByHeapIdx = (SD_FORWARD == sd)
        ? tpl.inShapes()
        : tpl.outShapes();

    const TShapeList &shapeList = shapeListByHeapIdx[fm.footprint.second];
    CL_BREAK_IF(1U != shapeList.size());
#endif

    const THeapIdent initHeap = (SD_FORWARD == sd)
        ? fm.matchedHeaps.front()
        : fm.matchedHeaps.back();

    const TShapeIdent initShape(initHeap, /* TODO */ 0);
    pInitShapes->insert(initShape);
}

void seekTemplateMatchInstances(
        TShapeIdentSet             *pInitShapes,
        MatchCtx                   &ctx,
        const OpTemplate           &tpl,
        const FootprintMatch       &fmInit,
        const TMetaOpSet           &metaOps,
        const TShapeIdent          &shIdent)
{

//# std::cout << "\n[function] seekTemplateMatchInstances:               #####" << std::endl;

    const ESearchDirection sd = tpl.searchDirection();
    SeekContext seekCtxInit(fmInit, metaOps, sd);

    // relocate object IDs in (from template generated) list of MOPs
    const SymHeap &sh = *heapByIdent(ctx.progState, shIdent.first);
    if (!relocObjsInMetaOps(&seekCtxInit.metaOpsToLookFor,
                seekCtxInit.objMapFromTpl, fmInit.props, sh)) {
//# std::cout << "!          seekTemplateMatchInstances: FAILED relocObjsInMetaOps" << std::endl;
        return;
    }
    // crawl the fixed-point
    // TODO: optimize to skip push/pack when traversing linear graphs
    std::stack<SeekContext> seekStack;
    seekStack.push(seekCtxInit);
    while (!seekStack.empty()) {
//# std::cout << "\n           seekTemplateMatchInstances: iterate through seekStack" << std::endl;
        const SeekContext seekCtxOrig = seekStack.top();
        seekStack.pop();

        const THeapIdent heapCurr = seekCtxOrig.heapCurrent;

        // collect the list of successor heaps (together with object mapping)
        THeapIdentList heapList;
        TObjectMapperList objMapList;
        collectNextHeaps(&heapList, &objMapList, heapCurr, ctx.progState, sd);
        const unsigned cnt = heapList.size();
        CL_BREAK_IF(cnt != objMapList.size());

        // iterate through successors
        for (unsigned idx = 0U; idx < cnt; ++idx) {
//std::cout << "\n           seekTemplateMatchInstances: iterate through successors idx=" << idx << std::endl;
            SeekContext seekCtx = seekCtxOrig;
            TMetaOpSet &metaOpsToLookFor = seekCtx.metaOpsToLookFor;
            FootprintMatch &fm = seekCtx.fm;

            const THeapIdent heapNext = heapList[idx];
            if (!insertOnce(seekCtx.seen, heapNext)) {
                // loop detected
//# std::cout << "--         seekTemplateMatchInstances:  loop detected = continue" << std::endl;
                continue;
            }

            // resolve the pair of heaps in program configuration
            const THeapIdent heap0 = (SD_FORWARD  == sd) ? heapCurr : heapNext;
            const THeapIdent heap1 = (SD_BACKWARD == sd) ? heapCurr : heapNext;

//# std::cout << "\nM \x1b[1m======== seekTemplateMatchInstances: ";
//# std::cout << "heap0=" << heap0 << " ---> heap1=" << heap1;
//# std::cout << " ========\x1b[0m\n" << std::endl;

            // resolve the current ID map
            const TObjectMapper &objMap = objMapList[idx];
            const SymHeap &sh = *heapByIdent(ctx.progState, heapNext);

            if (SD_BACKWARD == sd && !relocObjsInMetaOps(&metaOpsToLookFor,
                        objMap, fmInit.props, sh)) {
//# std::cout << "-          seekTemplateMatchInstances: FAILED SD_BACKWARD relocObjsInMetaOps() = continue" << std::endl;
                continue;
            }
            // INFO: we have heap1-relative mapping of MOPs

            if (!processDiffOf(&fm, &metaOpsToLookFor, ctx.progState,
                        heap0, heap1, sd)) {
                // failed to process the difference of the neighbouring heaps
//# std::cout << "-          seekTemplateMatchInstances:  FAILED to process the difference of the neighbouring heaps = continue" << std::endl;
                continue;
            }
            // INFO: single MOP matched and removed from metaOpsToLookFor

            // reflect the current traversal step in over-all mapping of objects
            // TODO: check if 2:2 IXI mapping is not a problem here
//# std::cout << "Template-prog mapping updated from: " << seekCtx.objMapFromTpl;
            seekCtx.objMapFromTpl.composite<D_LEFT_TO_RIGHT>(objMap);
//# std::cout << " to: " << seekCtx.objMapFromTpl << std::endl;

//# std::cout << "i          seekTemplateMatchInstances: remaining metaOpsToLookFor.size()=" << metaOpsToLookFor.size() << std::endl;

            if (metaOpsToLookFor.empty()) {
                // all template micro-ops matched!
                const EFootprintPort end = (SD_BACKWARD == sd)
                    ? FP_SRC
                    : FP_DST;
                fm.objMap[end] = seekCtx.objMapFromTpl;
                ctx.matchList.push_back(fm);

                insertInitShape(pInitShapes, tpl, fm);

                const THeapIdent src = fm.matchedHeaps.front();
                const THeapIdent dst = fm.matchedHeaps.back();
                CL_DEBUG("[ADT] template instance matched: tpl = " << tpl.name()
                        << "[" << fm.footprint.second << "]"
                        << ", src = " << src.first << "/" << src.second
                        << ", dst = " << dst.first << "/" << dst.second);

//# std::cout << "\n+++ \x1b[1m TEMPLATE INSTANCE MATCHED:\x1b[0m tpl = " << tpl.name() << "[" << fm.footprint.second << "]" << ", src = " << src.first << "/" << src.second << ", dst = " << dst.first << "/" << dst.second << std::endl << std::endl;

                continue;
            }
            else if (fm.matchedHeaps.empty()) {
                // we have not yet found the 1st insn implementing an operation
                const EFootprintPort beg = (SD_FORWARD == sd) ? FP_SRC : FP_DST;
                fm.objMap[beg] = seekCtx.objMapFromTpl;
//# std::cout << "i             seekTemplateMatchInstances: *** first implementing instruction NOT found" << std::endl;
            }

            if (SD_FORWARD == sd && !relocObjsInMetaOps(&metaOpsToLookFor,
                        objMap, fmInit.props, sh)) {
//# std::cout << "--         seekTemplateMatchInstances:  (SD_FORWARD && !relocObjsInMetaOps()) = continue" << std::endl;
                continue;
            }

//# std::cout << "i             seekTemplateMatchInstances: schedule the successor heap:" << heapNext << std::endl;
            // schedule the successor heap
            seekCtx.heapCurrent = heapNext;
            seekStack.push(seekCtx);
        } // iterate through successors LOOP
    }
//# std::cout << "           seekTemplateMatchInstances: normal return #####\n" << std::endl;
}

/// create the set of MOPs based on the template variant diff(fp.input, fp.output)
/// offsets in MOPs adjusted according to program (fm)
bool diffHeapsIfNeeded(
        TMetaOpSet                 *pMetaOps,
        const OpFootprint          &fp,
        const FootprintMatch       &fm)
{
    if (!pMetaOps->empty())
        // diff already computed
        return true;

//# std::cout << "[function] diffHeapsIfNeeded: === create TEMPLATE MetaOperation list" << std::endl;

    // time to diff the template
    if (!diffHeaps(pMetaOps, fp.input, fp.output)) {
        CL_BREAK_IF("AdtOp::diffHeaps() has failed");
//# std::cout << "- AdtOp::diffHeaps() has failed" << std::endl;
        return false;
    }

    // relocate next/prev/head offsets in MOPs according to program
    // INFO: does not change number of MOPs
    relocOffsetsInMetaOps(pMetaOps, fm);

    // TODO: move the test up
    if (pMetaOps->empty()) {
        CL_BREAK_IF("AdtOp::diffHeaps() returned an empty set");
//# std::cout << "- AdtOp::diffHeaps() returned an empty set" << std::endl;
        return false;
    }

//# std::cout << "T          diffHeapsIfNeeded: === TEMPLATE DIFF successfully computed ^ " << std::endl;
    // diff successfully computed
    return true;
}

// try to match single variant of template
void matchSingleFootprint(
        MatchCtx                   &ctx,
        const OpTemplate           &tpl,
        const OpFootprint          &fp,
        const TFootprintIdent      &fpIdent)
{
//# std::cout << "##################################################################" << std::endl;
//# std::cout << "[function] matchSingleFootprint:" << std::endl;
//# std::cout << "# fpIdent = " << fpIdent.first << "/" << fpIdent.second << std::endl;

    TMetaOpSet metaOps;
    TShapeIdentSet checkedShapes;

    BOOST_FOREACH(FixedPoint::TShapeSeq seq, ctx.shapeSeqs) {
        // for each shape sequence found earlier

        // check if the shapes in sequence and template are compatible
        if( ! tpl.isCompatible(*shapeByIdent(ctx.progState, seq.front()))) {
            // first shape in sequence and template types are not compatible
//# std::cout << "-          matchSingleFootprint: shapes and template not compatible" << std::endl;
            //continue; // FIXME: HACK uncomment when updated selectApplicableMatchesCore
        }

//# std::cout << "##########==============================================##########" << std::endl;
        // resolve shape sequence to search through
        const ESearchDirection sd = tpl.searchDirection();
        if (SD_FORWARD == sd)
            // reverse the sequence if searching _forward_
            std::reverse(seq.begin(), seq.end());

        // search anchor heap
        BOOST_FOREACH(const TShapeIdent &shIdent, seq) {
            // for each shape in sequence of consecutive shapes
            TMatchList matchList;
            if (!matchAnchorHeap(&matchList, ctx, tpl, fp, fpIdent, shIdent))
                // failed to match anchor heap
                continue;

            if (!insertOnce(checkedShapes, shIdent))
                // already checked as part of different shape sequence
                break;

            const THeapIdent heapIdent = shIdent.first;
            TM_DEBUG("found anchor heap: " << heapIdent.first
                    << "/" << heapIdent.second << ", seeking "
                    << tpl.name() << "[" << fpIdent.second << "] "
                    << ((SD_FORWARD == sd) ? "forward" : "backward")
                    << "...");

            BOOST_FOREACH(const FootprintMatch &fm, matchList) {
                // for each template variant
//# std::cout << "i         for each FootprintMatch AAAAAAAAAA" << std::endl;
                // compute metaop set iff not already computed
                if (!diffHeapsIfNeeded(&metaOps, fp, fm)) {
                    // non-recoverable error while computing diff of the footprint
//# std::cout << "! non-recoverable error while computing diff of the footprint" << std::endl;
                    return;
                }

                // find all template match instances using this anchor heap
                seekTemplateMatchInstances(&checkedShapes, ctx, tpl, fm,
                        metaOps, shIdent);
            }

            // now please continue with _another_ shape sequence!
            break;
        } // for each shape in shape sequence
    } // for each shape sequence

//# std::cout << "           matchSingleFootprint: normal return" << std::endl;
//# std::cout << "##################################################################" << std::endl;
//# std::cout << std::endl;
}

/// search for each template match in code
void matchTemplates(
        TMatchList                 *pDst,
        const OpCollection         &opCollection,
        const TProgState           &progState)
{
    MatchCtx ctx(*pDst, opCollection, progState);
    CL_DEBUG("[ADT] found " << ctx.shapeSeqs.size()
            << " container shape sequences");

//# std::cout <<  "\n[function] matchTemplates: Number of container shape sequences found = " << ctx.shapeSeqs.size() << "\n" << std::endl;

    const TTemplateIdx tplCnt = opCollection.size();

//# std::cout << "           matchTemplates: template count = " << tplCnt << std::endl;

    for (TTemplateIdx tplIdx = 0; tplIdx < tplCnt; ++tplIdx) {
        // for each template
        const OpTemplate &tpl = ctx.opCollection[tplIdx];
        const TFootprintIdx fpCnt = tpl.size();

//# std::cout << "\n           matchTemplates: template[" << tplIdx << "]=" << tpl.name() << ", size=" << fpCnt << "\n" << std::flush;

        for (TFootprintIdx fpIdx = 0; fpIdx < fpCnt; ++fpIdx) {
            // for each template variant
            const OpFootprint &fp = tpl[fpIdx];
            const TFootprintIdent fpIdent(tplIdx, fpIdx);
            TM_DEBUG("tpl = " << tpl.name() << "[" << fpIdx << "]"
                    ", looking for anchor heaps...");
            matchSingleFootprint(ctx, tpl, fp, fpIdent);
        }
    }
//# std::cout << "\n           matchTemplates: normal return\n" << std::endl;
}

/// if lastHeap is the last heap at location lastLoc, push lastLoc to *pDst
void pushBackIfLast(
        TInsnList                  *pDst,
        const TLocIdx               lastLoc,
        const THeapIdx              lastHeap,
        const TProgState           &progState)
{
    if (1U + lastHeap == progState[lastLoc].heapList.size())
        pDst->push_back(lastLoc);
}

typedef std::set<THeapIdent>                        THeapSet;

/// this assumes heapSet to be sorted by location idx, then by heap idx
void collectReplacedInsnsCore(
        TInsnList                  *pDst,
        const THeapSet             &heapSet,
        const TProgState           &progState)
{
    TLocIdx lastLoc = -1;
    THeapIdx lastHeap = -1;

    BOOST_FOREACH(const THeapIdent &heap, heapSet) {
        const TLocIdx currLoc = heap.first;
        if (-1 != lastLoc && lastLoc != currLoc) {
            // we have reached next location ==> was the last sequence correct?
            pushBackIfLast(pDst, lastLoc, lastHeap, progState);
            lastHeap = -1;
        }

        // check whether the current heap follows right after the previous one
        const THeapIdx currHeap = heap.second;
        lastLoc = (lastHeap + 1 == currHeap)
            ? currLoc
            : -1;

        // move to the next loc/heap pair
        lastHeap = currHeap;
    }

    if (-1 != lastLoc)
        // finally check the very last sequence
        pushBackIfLast(pDst, lastLoc, lastHeap, progState);
}

void collectReplacedInsns(
        TInsnListByTplIdx          *pDst,
        const TMatchList           &matchList,
        const TProgState           &progState)
{
    CL_BREAK_IF(!pDst->empty());

    // set of matched heaps per each template
    typedef std::vector<THeapSet>                   THeapSetByTpl;
    THeapSetByTpl heapSetByTpl;

    // go through all matched footprints and populate heapSetByTpl
    BOOST_FOREACH(const FootprintMatch &fm, matchList) {
        const THeapIdentSeq &heapList = fm.matchedHeaps;
        const TTemplateIdx tpl = fm.footprint./* tpl idx */first;
        if (heapSetByTpl.size() <= static_cast<unsigned>(tpl))
            heapSetByTpl.resize(1U + tpl);

        // skip the last element of the list while collecting the set
        THeapSet &heapSet = heapSetByTpl[tpl];
        std::copy(heapList.begin(), --heapList.end(),
                std::inserter(heapSet, heapSet.begin()));
    }

    // allocate space required for the result
    const TTemplateIdx tplCnt = heapSetByTpl.size();
    pDst->resize(tplCnt);

    // collect the set of replaced insns for each template
    TInsnListByTplIdx &dst = *pDst;
    for (TTemplateIdx tpl = 0; tpl < tplCnt; ++tpl)
        collectReplacedInsnsCore(&dst[tpl], heapSetByTpl[tpl], progState);
}

bool applyMatch(THeapSet *pHeapPool, const FootprintMatch &fm, const bool ro)
{
    // remove the last heap as it does not represent an insn to be replaced
    THeapIdentSeq matchedHeaps = fm.matchedHeaps;
    CL_BREAK_IF(matchedHeaps.size() < 2U);
    matchedHeaps.pop_back();

    if (ro) {
        // try it read-only
        BOOST_FOREACH(const THeapIdent &heap, matchedHeaps)
            if (!hasKey(pHeapPool, heap))
                // incomplete match!
                return false;
    }
    else {
        // remove all heaps used by this match
        BOOST_FOREACH(const THeapIdent &heap, matchedHeaps)
            pHeapPool->erase(heap);
    }

    // all OK!
    return true;
}

bool selectApplicableMatchesCore(
        THeapSet                   *pFailedSet,
        TMatchList                 *pMatchList,
        const TProgState           &progState)
{
    // collect instructions to be replaced
    const TMatchList &mlOrig = *pMatchList;
    TInsnListByTplIdx insnsToBeReplaced;
    collectReplacedInsns(&insnsToBeReplaced, mlOrig, progState);

    // collect heaps to be replaced
    THeapSet toReplace;
    BOOST_FOREACH(const TInsnList &insns, insnsToBeReplaced) {
        BOOST_FOREACH(const TLocIdx locIdx, insns) {
            const THeapIdx cntHeaps = progState[locIdx].heapList.size();
            for (THeapIdx heapIdx = 0; heapIdx < cntHeaps; ++heapIdx) {
                const THeapIdent heap(locIdx, heapIdx);
                toReplace.insert(heap);
            }
        }
    }

    // sort templates by the number of instructions they would replace
    typedef std::vector<TTemplateIdx>               TTplIdxList;
    typedef std::map<int /* prio */, TTplIdxList>   TTplIdxByPrio;
    TTplIdxByPrio tplIdxByPrio;
    const TTemplateIdx tplCnt = insnsToBeReplaced.size();
    for (TTemplateIdx tpl = 0; tpl < tplCnt; ++tpl) {
        const unsigned replCnt = insnsToBeReplaced[tpl].size();
        if (!replCnt)
            continue;

        const int prio = -replCnt;
        tplIdxByPrio[prio].push_back(tpl);
    }

    const unsigned matchCnt = mlOrig.size();

    // go through matches ordered by prio and select a set of applicable ones
    TMatchList mlSelected;
    BOOST_FOREACH(const TTplIdxByPrio::const_reference item, tplIdxByPrio)
    BOOST_FOREACH(const TTemplateIdx tpl, item.second) {
        // first check with matches are applicable
        std::vector<unsigned /* match idx */> picked;
        for (unsigned matchIdx = 0U; matchIdx < matchCnt; ++matchIdx) {
            const FootprintMatch &fm = mlOrig[matchIdx];
            if (fm.footprint./* tpl idx */first != tpl)
                continue;

            if (applyMatch(&toReplace, fm, /* ro */ true))
                picked.push_back(matchIdx);
        }

        // then remove the replaced heaps from the set and append the result
        BOOST_FOREACH(const unsigned matchIdx, picked) {
            const FootprintMatch &fm = mlOrig[matchIdx];
            applyMatch(&toReplace, fm, /* ro */ false);
            mlSelected.push_back(fm);
        }
    }

    if (toReplace.empty()) {
        // successful selection
        pMatchList->swap(mlSelected);
        return true;
    }
    else {
        // orphan heaps remained
        *pFailedSet = toReplace;
        return false;
    }
}

template <class TA, class TB>
bool intersects(const TA &aset, const TB &bset)
{
    BOOST_FOREACH(typename TA::const_reference item, aset)
        if (hasKey(bset, item))
            return true;

    // no intersection
    return false;
}

bool filterMatchList(
        TMatchList                 *pMatchList,
        const THeapSet             &blackList)
{
    TMatchList mlSelected;
    bool anyMatch = false;

    // iterate through matches
    BOOST_FOREACH(const FootprintMatch &fm, *pMatchList) {
        if (intersects(fm.matchedHeaps, blackList))
            anyMatch = true;
        else 
            mlSelected.push_back(fm);
    }

    pMatchList->swap(mlSelected);
    return anyMatch;
}

void selectApplicableMatches(
        TMatchList                 *pMatchList,
        const TProgState           &progState)
{
    THeapSet failedSet;
    while (!selectApplicableMatchesCore(&failedSet, pMatchList, progState)) {
        // remove the template matches causing the failure and try it again
        if (!filterMatchList(pMatchList, failedSet)) {
            CL_BREAK_IF("internal error in selectApplicableMatchesCore()");
            pMatchList->clear();
            break;
        }
    }

    CL_DEBUG("[ADT] count of applicable template match instances: "
            << pMatchList->size());
}

bool collectOpList(
        TOpList                    *pDst,
        const TMatchList           &matchList)
{
    CL_BREAK_IF(!pDst->empty());

    typedef std::map<TLocIdx, TOpIdx>               TIdxMap;
    TIdxMap idxMap;

    // go through matchList
    const TMatchIdx matchCnt = matchList.size();
    for (TMatchIdx matchIdx = 0; matchIdx < matchCnt; ++matchIdx) {
        const FootprintMatch &fm = matchList[matchIdx];

        // remove the last heap from fm.matchedHeaps (dst location vs. insn)
        THeapIdentSeq heapList(fm.matchedHeaps);
        CL_BREAK_IF(heapList.size() < 2);
        heapList.pop_back();

        // go through all locations (instructions to replace by this match)
        TOpIdx opIdx = -1;
        BOOST_FOREACH(const THeapIdent &heap, heapList) {
            const TLocIdx loc = heap.first;
            const TIdxMap::const_iterator it = idxMap.find(loc);

            if (idxMap.end() == it)
                continue;

            if (-1 != opIdx && opIdx != it->second) {
                TM_DEBUG("match collision detected by collectOpList()");
                return false;
            }

            // operation already assigned for this match
            opIdx = it->second;
        }

        if (-1 == opIdx) {
            // acquire a fresh operation idx for this match
            opIdx = pDst->size();
            pDst->push_back(TMatchIdxList());
        }

        // go through all locations (instructions to replace by this match)
        BOOST_FOREACH(const THeapIdent &heap, heapList) {
            const TLocIdx loc = heap.first;
            if (hasKey(idxMap, loc) && idxMap[loc] != opIdx)
                CL_BREAK_IF("internal error detected in collectOpList()");

            // remember the mapping for next operation
            idxMap[loc] = opIdx;
        }

        // append the current template match to the selected operation
        pDst->at(opIdx).push_back(matchIdx);
    }

    return /* success */ true;
}

} // namespace AdtOp
