/*
 * Copyright (C) 2013 Kamil Dudka <kdudka@redhat.com>
 *
 * This file is part of predator.
 *
 * predator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * predator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with predator.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"
#include "adt_op_meta.hh"

#include "adt_op_match.hh"          // for selectMappedObjByTs()
#include "fixed_point.hh"
#include "symtrace.hh"
#include "symseg.hh"                // for objMinLength()
#include "symutil.hh"               // for nextObj()

#ifndef NDEBUG
#   include "symplot.hh"
#endif

#include <cl/cl_msg.hh>

using FixedPoint::TObjectMapper;

namespace AdtOp {

bool debuggingHeapDiff = true;

#if 1
#define MO_DEBUG(msg) do {              \
    if (!AdtOp::debuggingHeapDiff)      \
        break;                          \
    CL_DEBUG(msg);                      \
} while (0)
#else
#define MO_DEBUG(msg) do {              \
    if (!AdtOp::debuggingHeapDiff)      \
        break;                          \
    std::cout << "DEBUG: " << msg << std::endl;      \
} while (0)
#endif

struct DiffHeapsCtx {
    TMetaOpSet                     &opSet;
    SymHeap                        &sh1;
    SymHeap                        &sh2;
    Trace::TIdMapper                idMap;

    DiffHeapsCtx(TMetaOpSet *pOpSet, SymHeap &sh1_, SymHeap &sh2_):
        opSet(*pOpSet),
        sh1(sh1_),
        sh2(sh2_)
    {
        resolveIdMapping(&idMap, sh1.traceNode(), sh2.traceNode());
        if (!idMap.isTrivial()) {
            MO_DEBUG("diffHeaps() operates on non-trivial map of object IDs");
        }
    }
};

/// select correct object from ObjList1 depending on mapped obj2 and direction in ts2
// INFO: single use function
// TODO: split into more functions and share them with other function
bool selectTargetObj(
        TObjList                   *pObjList1,
        DiffHeapsCtx               &ctx,
        const TObjId                obj2,
        const ETargetSpecifier      ts2)
{

//# std::cout << "[function] selectTargetObj:  obj2=" << obj2 << ", ts2=" << ts2 << std::endl;

    const unsigned cnt = pObjList1->size();
    switch (cnt) {
        case 0: // got no objects to select from
            MO_DEBUG("!  selectTargetObj: got no objects to select from");
            return false;

        case 1: // unambiguous ID mapping, we are done
            return true;

        case 2: // ambiguous ID mapping, continue
            break;

        default: // we do not support >=3
            MO_DEBUG("!  selectTargetObj: unsupported ObjList1.size>=3");
//abort();
            return false;
    }

    // 2:1 or 2:2 mapping supported

#if 1
//# std::cout << "i          selectTargetObj: AMBIGUOUS ID mapping, ";
//# std::cout << "cnt=" << cnt << "\t";
//# std::cout << "ObjList1=" << (*pObjList1) << std::endl;
#endif

    const EObjKind kind2 = ctx.sh2.objKind(obj2);
    switch(kind2) {
        //////////////////////////////////////////////////////////////////
        case OK_REGION: {
        //////////////////////////////////////////////////////////////////

//# std::cout << "i          selectTargetObj: {>1 obj}<-map->REGION ================" << std::endl;

        // INFO: only for src={dls0+,?}

        // INFO: this situation arises iff dls0+ removal AND concretization
        //       target region is the concretized one,
        // INFO2: ithere is variant dls0+ removal ONLY, too
        // problem: we need direction ts2!=TS_REGION to choose the right end of input sequence

        const TObjId fo = (*pObjList1)[0];
        const EObjKind kind1 = ctx.sh1.objKind(fo);
        const TObjId lo = (*pObjList1)[1];
        const EObjKind kind2 = ctx.sh1.objKind(lo);

// TODO: CHECK
        if(kind1 == OK_REGION && kind2 == OK_DLS && ctx.sh1.segMinLength(lo)==0) {
            // HACK possible? DLS0+ removal
            pObjList1->clear();
            pObjList1->push_back(fo);
            pObjList1->push_back(lo);   // signalization of DLS0 removal
            return true;
        }
        if(kind2 == OK_REGION && kind1 == OK_DLS && ctx.sh1.segMinLength(fo)==0) {
            // HACK possible? DLS0+ removal
            pObjList1->clear();
            pObjList1->push_back(lo);
            pObjList1->push_back(fo);   // signalization of DLS0 removal
            return true;
        }
// end

        if(kind1 != OK_DLS) {
//# std::cout << "-          selectTargetObj: FAIL first srcobj not DLS" << std::endl;
            return false;
        }
        if(kind2 != OK_DLS) {
//# std::cout << "-          selectTargetObj: FAIL last srcobj not DLS" << std::endl;
            return false;
        }

        const BindingOff bOff = ctx.sh1.segBinding(fo);
#if 0
        const TObjId obj1 = selectMappedObjByTs(ctx.sh1, bOff, *pObjList1, ts2);
#else
        // HACK: TODO we need to compute ts2={TS_LAST|TS_FIRST} depending on sibling dls
        // situation: mapping 2:2 full mapping
        // EXPERIMENT BUG HACK
        const TObjId obj1 = selectMappedObjByTs(ctx.sh1, bOff, *pObjList1, TS_LAST);
#endif
        if (OBJ_INVALID == obj1) {
            MO_DEBUG("selectMappedObjByTs() failed to resolve ambiguous mapping");
            return false;
        }

//# std::cout << "+          selectTargetObj: TRUE/REG obj1=" << obj1 << std::endl;
        pObjList1->clear();
        pObjList1->push_back(obj1);
        return true;
        }
        //////////////////////////////////////////////////////////////////
        case OK_DLS: {
        //////////////////////////////////////////////////////////////////

//std::cout << "i          selectTargetObj: {>1 obj}<-map->DLS ================" << std::endl;
            const BindingOff bOff = ctx.sh2.segBinding(obj2);
            // INFO: choose first in sequence (according to direction ts2)
            const TObjId obj1 = selectMappedObjByTs(ctx.sh1, bOff, *pObjList1, ts2);
            if (OBJ_INVALID == obj1) {
                MO_DEBUG("selectMappedObjByTs() failed to resolve ambiguous mapping");
                return false;
            }
//# std::cout << "+          selectTargetObj: TRUE/DLS obj1=" << obj1 << std::endl;
            pObjList1->clear();
            pObjList1->push_back(obj1);
            return true;
        }
        //////////////////////////////////////////////////////////////////
        default:
        //////////////////////////////////////////////////////////////////
//# std::cout << "-          selectTargetObj: UNSUPPORTED kind of obj2=" << obj2 << std::endl;
abort();
            break;
    } // switch kind2

    MO_DEBUG("selectMappedObjByTs() failed");
    return false;
} // selectTargetObj()

/// check if 2:2 IXI mapping, start with target object o2
bool isIXImap(DiffHeapsCtx &ctx, const TObjId o2)
{
    // First, get all srcobjects mapped to dst o2
    TObjList objList1;
    ctx.idMap.query<D_RIGHT_TO_LEFT>(&objList1, o2);
    if (2U != objList1.size()) {
        return false;   // not exactly 2 srcobjects
    }
    // we have V-map or IXI-map
    TObjSet objSet2; // all mapping targets
    // for all srcobjects accumulate targets
    BOOST_FOREACH(const TObjId &o1, objList1) {
        // get all targets mapped from o1
        TObjList objList2;
        ctx.idMap.query<D_LEFT_TO_RIGHT>(&objList2, o1);
        if (2U != objList2.size()) {
            return false;       // not exactly 2 targets
        }
        objSet2.insert(objList2[0]);
        objSet2.insert(objList2[1]);
    }

    if(objSet2.size() != 2U) {
        return false;   // not exactly 2 targets total
    }
    // 2:2 IXI-map
    return true;
}

/// select which mapped src objects/fields should be compared
/// returns single object in *pObjList1 iff next/prev offset, else all input
// TODO: solve all problem(s) if ambiguous mapping
bool selectObjsToCompare(
        TObjList                   *pObjList1,
        DiffHeapsCtx               &ctx,
        const FldHandle             fld2)
{
    const unsigned cnt = pObjList1->size();
//# std::cout << "[function] selectObjsToCompare: List1.size=" << cnt << std::endl;

    switch (cnt) {
        case 0:
//# std::cout << "-          selectObjsToCompare: return FALSE (ObjList1 empty)" << std::endl;
            return false;

        case 1: // 1:1 map
//# std::cout << "+          selectObjsToCompare: return TRUE  1:1" << std::endl;
            return true;

        case 2:
#if 1 // TEST, check next code
        case 3:
        case 4:
#endif
            break;

        default:
//# std::cout << "-          selectObjsToCompare: unsupported ObjList1.size=" << cnt << std::endl;
abort();
            return false;
    }

    // POSSIBILITIES:  2:1 V-map OR part of 2:2 IXI-map OR part of 3:2 W-map or ...
    // dst R: ( single R OR D1+ concretized ) and any number of DLS0+ removed
    // dst D: ( any number of R and D ) abstracted

    const TObjId obj2 = fld2.obj();
    CL_BREAK_IF(!ctx.sh2.isValid(obj2));

#if 1
//# std::cout << "i          selectObjsToCompare: obj2=" << obj2 << "\t ObjList1=";
//# std::cout << (*pObjList1) << std::endl;
#endif

    const EObjKind kind2 = ctx.sh2.objKind(obj2);
    if (OK_DLS != kind2) {
        //////////////////////////////////////////////////////////////////
        // INFO: more src objects mapped to single dst REGION
        //////////////////////////////////////////////////////////////////
        // it can be result of DLS0+ elimination and possibly DLS concretization

//# std::cout << "i          selectObjsToCompare: obj2 is REGION" << std::endl;

        // if source objects are compatible DLSs, we expect concretization op from dls+dls
        // INFO TODO: next is slightly modified copy of  [list]-->DLS  variant

        // gather the set of all objects in *pObjList1
        // AND check if all are DLS
        TObjSet objSet1;
        TObjId dls0plus=OBJ_INVALID;
        TObjId r=OBJ_INVALID;
        unsigned rnum=0;
        TObjId d=OBJ_INVALID;
        unsigned dnum=0;
        BOOST_FOREACH(const TObjId obj1, *pObjList1) {
            switch(ctx.sh1.objKind(obj1)) {
                case OK_REGION:
                    r = obj1;
                    ++rnum;
                    break;
                case OK_DLS:
                    ++dnum;
                    if(objMinLength(ctx.sh1, obj1) == 0) {
                        if(dls0plus==OBJ_INVALID)
                            dls0plus=obj1;      // first DLS0+
                        else {
                            MO_DEBUG("selectObjsToCompare() unsupported: too much DLS0+");
                            return false;
                        }
                    }
                    else {
                        d = obj1;
                    }
                    break;
                default:
                    MO_DEBUG("selectObjsToCompare() bad kind of src obj");
                    return false;
                    break;
            } // switch
            objSet1.insert(obj1);       // set of all src objects
        } // for each object in List1

//# std::cout << "i statistics: objList1 contains " << dnum << " DLS and " << rnum << " regions" << std::endl;
//# std::cout << "i          selectObjsToCompare: r=" << r << ", d=" << d << std::endl;

        if(rnum>1) {
            MO_DEBUG("selectObjsToCompare() BUG: more than 1 src REGION not possible");
            return false;
        }
        if(dls0plus==OBJ_INVALID) {
            MO_DEBUG("selectObjsToCompare() no DLS0+ not possible");
            return false;
        }

        // INFO: exactly one src object is DLS0+ (the abstracted one)
        // INFO: at most one REGION (mapped to dst REGION)
        // TODO: check if src objects compatible?

        const TOffset off = fld2.offset();
        const BindingOff bOff = ctx.sh1.segBinding(dls0plus);

        if (off != bOff.next && off != bOff.prev)
            // we require all other values to match along the path
            return true;

        //INFO: field offset off is equal to prev or next

// TODO: isolate to another function

        // get begin/end of the single sublist in ObjList1
        const TObjId beg = selectMappedObjByTs(ctx.sh1, bOff, *pObjList1, TS_FIRST);
        const TObjId end = selectMappedObjByTs(ctx.sh1, bOff, *pObjList1, TS_LAST);

        TObjId curr = beg;
        while (1U == objSet1.erase(curr) && !objSet1.empty()) {
            const TObjId next = nextObj(ctx.sh1, curr, bOff.next);
            if (!isOnHeap(ctx.sh1.objStorClass(next))) {
                // next object not on heap
                CL_DEBUG("TODO: diffFields() needs to be improved (non-heap obj1)");
                return false;
            }
            const TObjId prev = nextObj(ctx.sh1, next, bOff.prev);
            if (curr != prev)
                // wrong back-link
                return false;

            curr = next;
        }
        if (end != curr)
            return false;

        // INFO: all source objects linked as part of DLL

        TObjId obj1;
        if (off == bOff.prev)
            obj1 = beg;         // select first
        else if (off == bOff.next)
            obj1 = end;         // select last
        else
            return false;

#if 0
// EXPERIMENT:
        if(obj1!=dls0plus) {
//# std::cout << "HACK       selectObjsToCompare: obj1=dls0plus" << std::endl;
            obj1=dls0plus;      // choose DLS0+
        }
#endif

//# std::cout << "+++        selectObjsToCompare: return obj1=" << obj1 << std::endl;
//# std::cout << "//////////////////////////////////////////////////" << std::endl;

        pObjList1->clear();
        pObjList1->push_back(obj1);
        return true;
    } // [objList1] ---> obj2/REGION

    //////////////////////////////////////////////////////////////////
    // INFO: [objList1] ---> obj2/DLS
    //////////////////////////////////////////////////////////////////

//# std::cout << "i          selectObjsToCompare: obj2 is DLS " << std::endl;

    const TOffset off = fld2.offset();
    const BindingOff bOff = ctx.sh2.segBinding(obj2);
//# std::cout << "i          selectObjsToCompare: bOff=" << bOff << ", field.off=" << off << std::endl;
    if (off != bOff.next && off != bOff.prev)
        // we require non-poiner values to match along the path
        return true;

    //INFO: field offset off is equal to prev or next

    // get begin/end of the source chain
    const TObjId beg = selectMappedObjByTs(ctx.sh1, bOff, *pObjList1, TS_FIRST);
    const TObjId end = selectMappedObjByTs(ctx.sh1, bOff, *pObjList1, TS_LAST);

//# std::cout << "i          selectObjsToCompare: beg=" << beg << ", end=" << end << std::endl;

    // gather the set of all objects in *pObjList1
    TObjSet objSet1;
    BOOST_FOREACH(const TObjId obj1, *pObjList1)
        objSet1.insert(obj1);

    TObjId curr = beg;
    while (1U == objSet1.erase(curr) && !objSet1.empty()) {
        const TObjId next = nextObj(ctx.sh1, curr, bOff.next);
        if (!isOnHeap(ctx.sh1.objStorClass(next))) {
            // next object not on heap
            CL_DEBUG("diffFields() needs to be improved");
            return false;
        }

        const TObjId prev = nextObj(ctx.sh1, next, bOff.prev);
        if (curr != prev)
            // wrong back-link
            return false;

        curr = next;
    }
    if (end != curr)
        return false;

    TObjId obj1;
    if (off == bOff.prev)
        obj1 = beg;
    else if (off == bOff.next)
        obj1 = end;
    else
        return false;

// HACK: for DLS+DLS --> removal DLS0+ and concretization: IXI map2:2
//       (part 2:1 --> obj2/DLS)
//       we should compare removed dls with concretized region and
//       dls1+->dls0+ should be on the other end and compared to itself
// TODO: needs improvement (e.g for cnt>2)

if(cnt==2 && isIXImap(ctx,obj2)) {
    // TODO: FIXME D+D->D ?
    const EObjKind kind1b = ctx.sh1.objKind(beg);
    const EObjKind kind1e = ctx.sh1.objKind(end);
    if(kind1b==OK_DLS && kind1e==OK_DLS) {
        unsigned len = ctx.sh2.segMinLength(obj2);
        unsigned lenb = ctx.sh1.segMinLength(beg);
        unsigned lene = ctx.sh1.segMinLength(end);
        if(lenb==0 || lene==0) { // there is at least one DLS0+
            if(beg==obj2 && (lenb==0 || lenb-1==len)) { // unchanged id by mapping == concretized
                obj1=obj2;
//# std::cout << "HACK       selectObjsToCompare: IXI choose beg" << std::endl;
            } else if(end==obj2 && (lene==0 || lene-1==len)) {
                obj1=obj2;
//# std::cout << "HACK       selectObjsToCompare: IXI choose end" << std::endl;
            } else {
//# std::cout << "HACK       selectObjsToCompare: IXI fail?" << std::endl;
            }
        }
    }
}

//# std::cout << "+++        selectObjsToCompare: RETURN obj1=" << obj1 << std::endl;
//# std::cout << std::endl;

    pObjList1->clear();
    pObjList1->push_back(obj1);
    return true;
}

// check in sh2 if obj has some other object created by concretization
TObjId getConcretizationSibling(DiffHeapsCtx &ctx, const TObjId obj1, const TObjId obj2)
{
    // PRECONDITION: single concretization op in operation

    // parameter obj1 should be a DLS
    if (ctx.sh1.objKind(obj1) != OK_DLS)
        return OBJ_INVALID;

    // get all down mapped objects (obj1 -> exactly 2: DLS+R)
    TObjList tgtObjList2;
    ctx.idMap.query<D_LEFT_TO_RIGHT>(&tgtObjList2, obj1);
    if (tgtObjList2.size() != 2)
        return OBJ_INVALID;

    // check if one is obj2 and return the other
    for (int i = 0; i < 2; ++i) {
        if (tgtObjList2[i] == obj2)
            return tgtObjList2[!i];
    }

    return OBJ_INVALID;
}


// check if effect of operation  (obj1-map->obj2).field := value of fld2
// and ADD metaop
bool diffSetField(DiffHeapsCtx &ctx, const TObjId obj1, const FldHandle &fld2)
{
    // resolve val2 = value in field fld2
    const TValId val2 = fld2.value();
    const TObjType clt = fld2.type();
    const TOffset off = fld2.offset();

    // resolve target (pointer to tgtObj2)
    const TObjId tgtObj2 = ctx.sh2.objByAddr(val2);
    const TOffset tgtOff2 = ctx.sh2.valOffset(val2);
    const ETargetSpecifier tgtTs2 = ctx.sh2.targetSpec(val2);

    const TObjId obj2 = fld2.obj();
//# std::cout << "[function]\x1b[1m diffSetField:\x1b[0m obj1=" << obj1;
//# std::cout << "  sh2/o#" << obj2 << ".field#"<<fld2.fieldId()<<"-->tgtObj2=" << tgtObj2;
//# std::cout << " offset2=" << tgtOff2 << " tgtTs2=" << tgtTs2 << std::endl;

    const EValueTarget vt2 = ctx.sh2.valTarget(val2);
    switch (vt2) {
        case VT_OBJECT: // address of object
        case VT_CUSTOM: // other value
            break;

        case VT_UNKNOWN:
            return true;

        default:
            CL_BREAK_IF("diffSetField() does not support non-pointer fields");
            return false;
    }

    // check target object mapping
    TObjList tgtObjList1;
    ctx.idMap.query<D_RIGHT_TO_LEFT>(&tgtObjList1, tgtObj2);
    TObjList tgtObjList1all=tgtObjList1; // HACK

    // INFO: tgtObjList1 contains all srcobjects mapped to tgtObj2
    // if not 1:1, we should choose right end-object of sequence
    if (!selectTargetObj(&tgtObjList1, ctx, tgtObj2, tgtTs2)) {
        MO_DEBUG("selectTargetObj() failed to resolve ambiguous ID mapping");
        return false;
    }

#if 1
    bool wasDLS0removal=false;
    TObjId dls0removed=OBJ_INVALID;
    if(tgtObjList1.size()==2) { // HACK: signalization of special case
        dls0removed=tgtObjList1[1];
        tgtObjList1.pop_back();
        wasDLS0removal=true;
    }
#endif

    // INFO: tgtObjList1 contains single srcobject mapped to tgtObj2
    CL_BREAK_IF(tgtObjList1.size()!=1);

    const TObjId tgtObj1 = tgtObjList1.front(); // src target

    if (ctx.sh1.isValid(obj1)) {        // valid obj1 => valid field?
//# std::cout << "i          diffSetField: obj1 kind=" << ctx.sh1.objKind(obj1) << std::endl;
//# std::cout << "i          diffSetField: obj2 kind=" << ctx.sh2.objKind(obj2) << std::endl;

        // resolve val1
        const FldHandle fld1(ctx.sh1, obj1, clt, off);
        const TValId val1 = fld1.value();
        if (val1 == val2)
            return true;        // identical values

        const EValueTarget vt1 = ctx.sh1.valTarget(val1);
        switch (vt1) {
            case VT_UNKNOWN:
                break;

            case VT_OBJECT:
                // INFO: val1 is address of object

                if (ctx.sh1.objByAddr(val1) != tgtObj1)
                    break;
                // INFO: the same val1 and val2 = same target objects (modulo mapping)

                if (ctx.sh1.valOffset(val1) != tgtOff2)
                    break;
                // INFO: the same offset in target

//TODO: BUG? 4 possibilities of target type: RR,DR,RD,DD
                if (/* XXX */ TS_REGION != tgtTs2
                        && ctx.sh1.targetSpec(val1) != tgtTs2)
                    break;
                // tgt2 is region (RR=OK, DR=means target concretization?)
                // OR matching target-spec for DLS (DD=OK, RD always breaks)

#if 0
// TODO-CHECK: is it REALLY the only possibility?
if(TS_REGION == tgtTs2 && ctx.sh1.targetSpec(val1) != TS_REGION) { // DR
// INFO: field value change only due to concretization of target
//# std::cout << "XXXXXXXXXXXXXXXX DR: optimized concretization of sh1/o1#" << tgtObj1 << std::endl;
//    break;    // we can break, add MOP and then remove MOP, but why?
}
#endif
                // nothing changed actually
                return true;

            case VT_CUSTOM:
                MO_DEBUG("diffSetField() ignores change of a non-pointer field");
                return true;

            default:
                CL_BREAK_IF("unhandled value target in diffSetField()");
                return false;
        } // switch

/////////////////////////////////////////////////////////////////////////////
// FILTER1
/////////////////////////////////////////////////////////////////////////////
// INFO: we do not add MOPs representing internal heap restructuralization
// check if target is concretization sibling and prev/next offset of
// field => return true ==> ignore artificial change due to concretization
// maybe check other fields? should be equal?
// TODO: consider deduction of MOPs from symexec of instructions.

        TObjId obj2sibling = getConcretizationSibling(ctx,obj1,obj2);
        if(obj2sibling!=OBJ_INVALID) {
            // TODO: is A-map always the concretization?
            // concretization
//# std::cout << "i          diffSetField: sibling of obj2="<< obj2 <<" is obj#" << obj2sibling << std::endl;
            // HACK? we know last program-writen field in sh2
            TFldId lastmod = ctx.sh2.getLastModField();
//# std::cout << "i          diffSetField: sh2.LastModField=" << lastmod << std::endl;

            if(lastmod != fld2.fieldId()) {
                // field is not the last writen by program instruction
                // so the operation is artificial
//# std::cout << "-          diffSetField: field#"<< fld2.fieldId() <<" concretization change IGNORED" << std::endl;
                return true;
            }
            // continue by MOP ADD

// TODO: remove next code 
#if 0
            EObjKind obj2kind = ctx.sh2.objKind(obj2);
            EObjKind obj2siblingkind = ctx.sh2.objKind(obj2sibling);
#endif
        } // if has sibling
/////////////////////////////////////////////////////////////////////////////
// FILTER2
/////////////////////////////////////////////////////////////////////////////
// If there was DLS0 removal, we need to remove all write ops except the real
// one
        if(wasDLS0removal) {
            // HACK? we know last program-writen field in sh2
            TFldId lastmod = ctx.sh2.getLastModField();
//# std::cout << "i          diffSetField: sh2.LastModField=" << lastmod << std::endl;
            if(lastmod != fld2.fieldId()) {
                // field is not the last writen by program instruction
                // so the operation is artificial
//# std::cout << "-          diffSetField: field#"<< fld2.fieldId() <<" DLS0 removal change IGNORED" << std::endl;
                return true;
            }
            // continue by MOP ADD
        } // if wasDLS0removal

/////////////////////////////////////////////////////////////////////////////
// FILTER3
/////////////////////////////////////////////////////////////////////////////
// If there was abstraction of target, we need to consider correct src target
// equal to correct end of mapped dst DLS
// TODO: it should be solved before this  (V-map always means abstraction?)

        if(!wasDLS0removal /*antiHACK*/ && tgtObj1!=tgtObj2) {
            // HACK? target mapping differs ==> abstraction to tgtObj2

// TODO: check if target really abstracted out! ###
// INFO it seems not to collide with NULL-DLL

            // FIXME check if correct
//# std::cout << "-          diffSetField: tgtObj1!=tgtObj2 Abstraction MOP IGNORED" << std::endl;
            return true; // ignore false MOP
            // continue by MOP ADD
        } // if wasDLS0removal

/////////////////////////////////////////////////////////////////////////////
    } else { // not valid obj1
//# std::cout << "i          diffSetField: INVALID obj1=" << obj1 << std::endl;
    }

if(obj2!=obj1) {
//MAYBE BUG: operation on concretized object on the inside part
//# std::cout << "W          diffSetField: (obj2=fld2.obj())!=obj1 CHECK this" << std::endl;
}

// TODO FIXME: consider use target objid?
// we lose which concretized objects from DLS are written to (at least in the
// case of cyclic DLL

    // insert meta-operation
    /*const*/ TObjId obj = (ctx.sh1.isValid(obj1)) ? obj1 : obj2;

if(wasDLS0removal) {
    // HACK
    // if DLS0 removed, the only possibility is write through it
    // TODO: what if tgtObj is the same object ?
//# std::cout << "i          diffSetField: MOP write to removed DLS0=" << dls0removed << std::endl;
    obj=dls0removed;

//TODO filtr falesnych operaci pro dls0 removal v dalsim pripadu

}


    const MetaOperation moSet(MO_SET, obj, off, tgtObj2, tgtOff2, tgtTs2, obj2);
//# std::cout << "+          diffSetField: ADD " << moSet << std::endl;
    ctx.opSet.insert(moSet);
    return true;
}

bool diffUnsetField(DiffHeapsCtx &ctx, const FldHandle &fld1, const TObjId obj2)
{

//# std::cout << "[function] diffUnsetField: obj2=" << obj2 << std::endl;

    // resolve val1
    const TValId val1 = fld1.value();
    const EValueTarget vt1 = ctx.sh1.valTarget(val1);
    switch (vt1) {
        case VT_OBJECT:
        case VT_CUSTOM:
            break;

        case VT_UNKNOWN:
            return true;

        default:
            CL_BREAK_IF("diffUnsetField() does not support non-pointer fields");
            return false;
    }

    // resolve val2
    const TObjType clt = fld1.type();
    const TOffset off = fld1.offset();
    const FldHandle fld2(ctx.sh2, obj2, clt, off);
    const TValId val2 = fld2.value();
    if (val1 == val2)
        // identical values
        return true;

    const EValueTarget vt2 = ctx.sh2.valTarget(val2);
    if (VT_UNKNOWN != vt2)
        // this is NOT an "unset" operation
        return true;

    // check object mapping
    const TObjId obj1 = fld1.obj();
    if (obj1 != obj2) {
        // FIXME: we blindly assume that our abstraction threw away the value
        MO_DEBUG("diffUnsetField() does not support non-trivial object map");
        return true;
    }

    // insert meta-operation
    const MetaOperation moUnset(MO_UNSET, obj1, off);
//# std::cout << "+          diffUnsetField: ADD " << moUnset << std::endl;
    ctx.opSet.insert(moUnset);
    return true;
}

/// compare all fields of all? mapped objects
bool diffFields(DiffHeapsCtx &ctx, const TObjList &objList1, const TObjId obj2)
{
//# std::cout << "[function] diffFields:  obj2=" << obj2 << std::endl;

    BOOST_FOREACH(const TObjId obj1, objList1) {
        if (!ctx.sh1.isValid(obj1)) {
            // skip invalid srcobjects TODO: isnt it a problem?
//# std::cout << "i          diffFields: invalid obj1=" << obj1 << std::endl;
            continue;
        }

        const TSizeRange size1 = ctx.sh1.objSize(obj1);
        const TSizeRange size2 = ctx.sh2.objSize(obj2);
        if (size1 != size2) {
            CL_BREAK_IF("object size mismatch in diffFields()");
            return false;       // different size can not match
        }

        FldList fldList1;
        ctx.sh1.gatherLiveFields(fldList1, obj1);
        BOOST_FOREACH(const FldHandle &fld1, fldList1)
            if (!diffUnsetField(ctx, fld1, obj2)) {
//# std::cout << "-          diffFields: FALSE (!diffUnsetField(ctx, fld1, obj2)) " << std::endl;
                return false;
            }
    }

    FldList fldList2;
    ctx.sh2.gatherLiveFields(fldList2, obj2);
    BOOST_FOREACH(const FldHandle &fld2, fldList2) {
        // for each live field fld2 of obj2
        TObjList selectedObjs1(objList1);
        if (!selectObjsToCompare(&selectedObjs1, ctx, fld2)) {
            MO_DEBUG("selectObjsToCompare() has failed");
            return false;
        }

        BOOST_FOREACH(const TObjId obj1, selectedObjs1) {
            if (!diffSetField(ctx, obj1, fld2)) {
//# std::cout << "_          diffFields: FALSE (!diffSetField(ctx, obj1, fld2)) " << std::endl;
                return false;
            }
        }
    }

    // fields diffed successfully!
//std::cout << "_ fields diffed successfully! " << std::endl;
    return true;
}

// check if objList contains single DLS only, return binding
bool findSingleDls(BindingOff *pDst, const SymHeap &sh, const TObjList &objList)
{
    unsigned dlsCount = 0U;

    BOOST_FOREACH(const TObjId obj, objList) {
        const EObjKind kind = sh.objKind(obj);
        switch (kind) {
            case OK_REGION:
                continue;

            case OK_DLS:
                break;

            default:
                CL_BREAK_IF("unexpected kind of object in findSingleDls()");
//# std::cout << "!         findSingleDls: unexpected kind of object in findSingleDls()" << std::endl;
                return false;
        }

        ++dlsCount;
        *pDst = sh.segBinding(obj);
    }

    return (1U == dlsCount);
}


/// check if metaop is due to concretization (=not REAL op)
/// FIXME: it will be better not to add them
bool isConcretizationOp(DiffHeapsCtx &ctx, const MetaOperation &mo)
{

//std::cout << "[function] isConcretizationOp: " << mo << std::endl;

    if (MO_SET != mo.code)
        // only MO_SET ops are handled for now
        return false;

    // get all srcobjects mapped to dst mo.obj
    TObjList objList1;
    ctx.idMap.query<D_RIGHT_TO_LEFT>(&objList1, mo.obj);

    if (0U == objList1.size()) {
//# std::cout << "WARNING zero srcobjects = not concretization" << std::endl;
        return false;
    }

    if (1U != objList1.size()) {
#if 1
        // not a concretization (at least not the simple case)
        return false;
#else

        ///////////////////////////////////////////////////////////////////////////
        // concretization check for non-trivial case od 2:2 map

//# std::cout << "i CONCRETIZATION   srcsize=" << objList1.size() << std::endl;

        if (2U != objList1.size()) {
            // TODO: is it possible to have more objects mapped?
//# std::cout << "ERROR more than 2 srcobjects = not implemented" << std::endl;
abort();
            return false;
        }

        // 2:1 V-map or 2:2 IXI-map

        if (!isIXImap(ctx,mo.obj)) {
//# std::cout << "ERROR not 2:2 IXI map = not implemented" << std::endl;
abort();
            return false;
        }

        // INFO: IXI map

        const EObjKind kind2 = ctx.sh2.objKind(mo.obj);

        int dls0plus = 0;
        TObjId srcdls0 = OBJ_INVALID;
        TObjId srcdls1 = OBJ_INVALID;
        BOOST_FOREACH(const TObjId &s, objList1) {
            const EObjKind k = ctx.sh1.objKind(s);
            if(k != OK_DLS) {   // not supported
//# std::cout << "IXI ERROR: non DLS srcobject = not implemented" << std::endl;
                return false;
            }
            if(ctx.sh1.segMinLength(s) == 0) {
                ++dls0plus;
                srcdls0 = s;
            }
            if(ctx.sh1.segMinLength(s) > 0) {
                srcdls1 = s;
            }
        }
        if (dls0plus != 1) {
//# std::cout << "IXI ERROR: not exactly one DLS0 srcobject = not implemented" << std::endl;
            return false;
        }

        // INFO: both srcobjects are DLS and one DLS0+ present
        //       it is concretization together with DLS0+ removal
        TObjList objList2;
        const TObjId obj1a = objList1.front();
        ctx.idMap.query<D_LEFT_TO_RIGHT>(&objList2, obj1a);

        int regnum = 0;
        TObjId tgtdls0 = OBJ_INVALID;
        TObjId tgtreg = OBJ_INVALID;
        BOOST_FOREACH(const TObjId &t, objList2) {
            const EObjKind k = ctx.sh2.objKind(t);
            switch(k) {
                case OK_DLS:
                    tgtdls0 = t;
                    break;
                case OK_REGION:
                    tgtreg = t;
                    ++regnum;
                    break;
                default:
//# std::cout << "IXI ERROR: bad kind of target obj = not implemented" << std::endl;
                    return false;
                    break;
            }
        } // foreach target t
        if (regnum != 1) {
//# std::cout << "IXI ERROR: not exactly one REGION target object = not implemented" << std::endl;
            return false;
        }
        // INFO: IXI map:  srcdls0+, srcdls1+ ---> tgtreg, tgtdls0+

//# std::cout << "IXI ERROR: (dls0+,dls1+)--->obj    = not implemented" << std::endl;
abort();
//////////////////////////////////////////////////////////////////
// TODO solve mapping problem - see selectTargetObj, too
//        return true;
//////////////////////////////////////////////////////////////////


        return false;
#endif
    } // objList1.size() != 1   not 1:(N>=1) map

    // single srcobject (always should be DLS):
    const TObjId obj1 = objList1.front();
    // gather dstobjects
    TObjList objList2;
    ctx.idMap.query<D_LEFT_TO_RIGHT>(&objList2, obj1);
    if (2U != objList2.size()) {
        // not a concretization (at least not the simple case)
//# std::cout << "ERROR: not a 1:2 concretization = not implemented" << std::endl;
        return false;
    }

    const TObjList::const_iterator beg2 = objList2.begin();
    const TObjList::const_iterator end2 = objList2.end();

    if (end2 == std::find(beg2, end2, mo.obj))
        // mo.obj not on the list of dst objects
        return false;

    if (end2 == std::find(beg2, end2, mo.tgtObj))
        // mo.tgtObj not on the list of dst objects
        return false;

    // INFO: both MOP objects found in dstobjects
    //       (e.g. R<->DLS: MOP can be result of concretization)

    BindingOff bOff;
    if (!findSingleDls(&bOff, ctx.sh2, objList2))
        // not a DLS + REG pair of objects = no concretization
        return false;

    if (mo.tgtOff != bOff.head)
        // target offset mismatch
        return false;

return false; // cyclic HACK -- FIXME check field modification

    switch (mo.tgtTs) {
        case TS_FIRST:  // tgtObj is DLS, normal direction R->DLS
            return mo.off == bOff.next; // TRUE if access inside R<->DLS

        case TS_LAST:   // tgtObj is DLS, reverse direction DLS<-R
            return mo.off == bOff.prev; // TRUE if access inside R<->DLS

        case TS_REGION: // tgtObj is R
            return mo.off == bOff.next
                || mo.off == bOff.prev;

        default:
            CL_BREAK_IF("invalid target specifier in isConcretizationOp()");
            return false;
    }
}

void dropConcretizationOps(DiffHeapsCtx &ctx)
{

return;
// TODO: this function is NOT NEEDED - See hack to filter the concretization metaops
//# std::cout << "[function] dropConcretizationOps:   remove metaops produced by concretization " << std::endl;

    TMetaOpSet result;

    BOOST_FOREACH(const MetaOperation &mo, ctx.opSet) {
        if (isConcretizationOp(ctx, mo)) {
//# std::cout << "           dropConcretizationOps: REMOVE " << mo << std::endl;
            continue;
        }

        result.insert(mo);
    }

//# std::cout << "           dropConcretizationOps: DONE " << std::endl;

    ctx.opSet.swap(result);
}

// compare two heaps and produce (set of) metaops
bool diffHeaps(TMetaOpSet *pDst, const SymHeap &sh1, const SymHeap &sh2, int loc)
{

//# std::cout << "[function] diffHeaps: " << std::endl;

    DiffHeapsCtx ctx(pDst,
            const_cast<SymHeap &>(sh1),
            const_cast<SymHeap &>(sh2));

//# std::cout << "i MAP:" << ctx.idMap << std::endl;

    TObjList objList2;
#if 0
    ctx.sh2.gatherObjects(objList2);
#else
    ctx.sh2.gatherObjects(objList2,isOnHeap);   // BUG? TODO: isOnHeap was missing!
#endif
    BOOST_FOREACH(const TObjId obj2, objList2) {

//# std::cout << std::endl;
//# std::cout << "[]         diffHeaps: obj2=" << obj2 << "     =====DIFF=====" << std::endl;

        TObjList objList1;
        ctx.idMap.query<D_RIGHT_TO_LEFT>(&objList1, obj2);

        if (objList1.empty() || !ctx.sh1.isValid(objList1.front())) {
            // ALLOC:  nothing -> obj2
            if (OK_REGION != ctx.sh2.objKind(obj2)) {
                // only regions are supported with MO_ALLOC for now
//# std::cout << "-          diffHeaps: only regions are supported with MO_ALLOC for now" << std::endl;
                return false;
            }

            const MetaOperation moAlloc(MO_ALLOC, obj2);
//# std::cout << "i          diffHeaps: ADD ALLOC " << moAlloc << std::endl;
            ctx.opSet.insert(moAlloc);
        } // if ALLOC

        if (!diffFields(ctx, objList1, obj2)) {
//# std::cout << "-          diffHeaps: !diffFields(ctx, objList1, obj2), return FALSE" << std::endl;
            return false;
        }
    }

    TObjList objList1;
#if 0
    ctx.sh1.gatherObjects(objList1);
#else
    ctx.sh1.gatherObjects(objList1,isOnHeap);   // BUG? TODO: isOnHeap was missing!
#endif
    BOOST_FOREACH(const TObjId obj1, objList1) {
//# std::cout << "?          diffHeaps: obj1=" << obj1 << std::endl;
        TObjList objList2;
        ctx.idMap.query<D_LEFT_TO_RIGHT>(&objList2, obj1);
        if (objList2.empty())                   // obj1 removed
            objList2.push_back(OBJ_INVALID);

        BOOST_FOREACH(const TObjId obj2, objList2) {
            if (!ctx.sh2.isValid(obj2)) {
                // FREE:  obj1 -> nothing
                if (OK_REGION != ctx.sh2.objKind(obj2)) {
                    // only regions are supported with MO_FREE for now
//# std::cout << "-          diffHeaps: only regions are supported with MO_FREE for now" << std::endl;
                    return false;
                }

                const MetaOperation moFree(MO_FREE, obj1);
//# std::cout << "i          diffHeaps: ADD FREE " << moFree << std::endl;
                ctx.opSet.insert(moFree);
            } // IF FREE
        }
    }

    if (!ctx.idMap.isTrivial()) {
        dropConcretizationOps(ctx);
    }

//# std::cout << std::endl;
//# std::cout << "i          diffHeaps: sh1 LOCATION = " << loc << std::endl;
//# std::cout << "+          diffHeaps: RETURN - heaps diffed successfully with result = " << std::endl;
//# std::cout << ctx.opSet;
//# std::cout << std::endl;

    // heaps diffed successfully!
    return true;
}

} // namespace AdtOp

#ifndef NDEBUG
void sl_dump(const AdtOp::DiffHeapsCtx &ctx)
{
    plotHeap(ctx.sh1, "diffHeaps");
    plotHeap(ctx.sh2, "diffHeaps");
}
#endif
