/*
 * Copyright (C) 2013 Kamil Dudka <kdudka@redhat.com>
 *
 * This file is part of predator.
 *
 * predator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * predator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with predator.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef H_GUARD_ADT_OP_META_H
#define H_GUARD_ADT_OP_META_H

#include "util.hh"                          // for RETURN_IF_COMPARED
#include "symheap.hh"
#include <iostream>                             // std::cout etc

namespace AdtOp {

//// enumeration of supported meta-operations (only pointer operations for now)
enum EMetaOperation {
    MO_INVALID = 0,
    MO_ALLOC,                               /// a new object is introduced
    MO_FREE,                                /// an existing object disappears
    MO_SET,                                 /// a value is written to a field
    MO_UNSET                                /// a tracked field is invalidated
};

//// description of a meta-operation (only pointer operations for now)
struct MetaOperation {
    EMetaOperation      code;               /// kind of meta-operation
    TObjId              obj;                /// object ID we operate on
    TOffset             off;                /// offset we write at (set/unset)
    TObjId              tgtObj;             /// object we take address of (set)
    TOffset             tgtOff;             /// tgt offset of the address (set)
    ETargetSpecifier    tgtTs;              /// tgt specifier of the addr (set)
    TObjId              obj2;               /// HACK: object ID we operate on next heap

    MetaOperation(
            const EMetaOperation    code_   = MO_INVALID,
            const TObjId            obj_    = OBJ_INVALID,
            const TOffset           off_    = 0,
            const TObjId            tgtObj_ = OBJ_INVALID,
            const TOffset           tgtOff_ = 0,
            const ETargetSpecifier  tgtTs_  = TS_INVALID,
            const TObjId            obj2_   = OBJ_INVALID):
        code    (code_),
        obj     (obj_),
        off     (off_),
        tgtObj  (tgtObj_),
        tgtOff  (tgtOff_),
        tgtTs   (tgtTs_),
        obj2    (obj2_)
    {
    }
};

/// print MetaOperation
inline std::ostream & operator<< (std::ostream &s, const MetaOperation &mo) {
    return s << "MetaOperation(" <<
    "code=" << mo.code <<
    ", obj=" << mo.obj <<
    "/" << mo.obj2 <<
    ", off=" << mo.off <<
    ", tgtObj=" << mo.tgtObj <<
    ", tgtOff=" << mo.tgtOff <<
    ", tgtTs=" << mo.tgtTs << ")";

}

#define MOP_COMPARE(a, b, member) do {   \
    if (a.member != b.member)        \
        return false;                \
} while (0)

/// comparison
inline bool operator==(const MetaOperation &a, const MetaOperation &b)
{
    MOP_COMPARE(a, b, code);
    MOP_COMPARE(a, b, obj);
    MOP_COMPARE(a, b, off);
    MOP_COMPARE(a, b, tgtObj);
    MOP_COMPARE(a, b, tgtOff);
    MOP_COMPARE(a, b, tgtTs);
    return true; // all the items are equal
}
#undef MOP_COMPARE

inline bool operator!=(const MetaOperation &a, const MetaOperation &b) {
    return ! operator==(a,b);
}

/// required to be defined by the implementation of std::set
inline bool operator<(const MetaOperation &a, const MetaOperation &b)
{
    // compare lexicographically
    RETURN_IF_COMPARED(a, b, code);
    RETURN_IF_COMPARED(a, b, obj);
    RETURN_IF_COMPARED(a, b, off);
    RETURN_IF_COMPARED(a, b, tgtObj);
    RETURN_IF_COMPARED(a, b, tgtOff);
    RETURN_IF_COMPARED(a, b, tgtTs);

    // the items are incomparable
    return false;
}

/// set of meta-operations (which should be independent on each other)
typedef std::set<MetaOperation>                     TMetaOpSet;

/// print set of metaops
inline std::ostream & operator<< (std::ostream &s, const TMetaOpSet &mos) {
    s << "TMetaOpSet{";
    TMetaOpSet::iterator b=mos.begin();
    TMetaOpSet::iterator e=mos.end();
    if(b!=e) s << std::endl;
    else     s << " ";
    for(; b!=e; b++) {
        s << "  " << (*b) << std::endl;
    }
    return s << "}" << std::endl;
}


/// compare two heaps (sh1 and its successor sh2), create set of metaops
bool diffHeaps(TMetaOpSet *pDst, const SymHeap &sh1, const SymHeap &sh2,
int input_sh_location=-1);

/// print list of objects TODO: move to symheap.hh
inline std::ostream & operator<< (std::ostream &s, const TObjList &l) {
    s << "TObjList[ ";
    TObjList::const_iterator b=l.cbegin();
    TObjList::const_iterator e=l.cend();
    for(; b!=e; b++) {
        s << (*b) << ", ";
    }
    return s << "]";
}

} // namespace AdtOp

#endif /* H_GUARD_ADT_OP_META_H */
