/*
 * Copyright (C) 2013 Kamil Dudka <kdudka@redhat.com>
 *
 * This file is part of predator.
 *
 * predator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * predator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with predator.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "cont_shape.hh"

#include "symdiscover.hh"
#include "symstate.hh"
#include "symutil.hh"

extern int debug_location;
unsigned debug_smg;     // HACK for location info

namespace ContShape {

bool debuggingEnabled = true;   /// set to true if needed

class ApparentShapeDetector {
    public:
        ApparentShapeDetector(SymHeap &sh, TShapeList &dstArray):
            sh_(sh),
            dstArray_(dstArray)
        {
        }

        bool probeEntry(const TObjId obj, const ShapeProps &props);

    private:
        typedef std::map<ShapeProps, TObjSet>       TObjsByProps;

        SymHeap                    &sh_;
        TShapeList                 &dstArray_;
        TObjsByProps                objsByProps_;

        bool isFreeEnt(const TObjId obj, const ShapeProps &props);
};

bool ApparentShapeDetector::isFreeEnt(const TObjId obj, const ShapeProps &props)
{
    const TObjsByProps::const_iterator it = objsByProps_.find(props);
    if (it == objsByProps_.end())
        return true;

    const TObjSet &taken = it->second;
    return !hasKey(taken, obj);
}

// TODO: split into more functions: a) search for SHAPEPROTOTYPE,
//                                  b) check terminators,
//                                  c) normalize?
bool ApparentShapeDetector::probeEntry(const TObjId obj, const ShapeProps &props)
{
//std::cout << "[function] ApparentShapeDetector::probeEntry: obj=" << obj << " props=" << props << std::endl;

    if (!this->isFreeEnt(obj, props))
        return false;

    TObjSet seen;

    // seek beg
    TObjId beg = obj;
    for (;;) {
        seen.insert(beg);
        TObjId prev = nextObj(sh_, beg, props.bOff.prev);
        TObjId prevNext;
        if (!canMergeObjWithNextObj(sh_, prev, props, &prevNext))
            break;
        if (prevNext!=beg) // there is a link to another object
            break;

        if (hasKey(seen, prev))
            // loop detected (we intentionally do not use insertOnce() here)
            break;

        beg = prev;
    }

    // seek end
    TObjId end = obj;
    for (;;) {
        TObjId next;
        if (!canMergeObjWithNextObj(sh_, end, props, &next))
            break;

        if (hasKey(seen, next))
            // loop detected (we intentionally do not use insertOnce() here)
            break;

        end = next;
        seen.insert(end);
    }

    // found non-empty DLL segment from begin object to end object, possibly begin==end

    // check the length
    const unsigned len = seen.size();

    // check the end-points
    const TValId valPrev = valOfPtr(sh_, beg, props.bOff.prev);
    const TValId valNext = valOfPtr(sh_, end, props.bOff.next);

    EShapeType shapeType = SHAPE_UNKNOWN;
    int variant = ((valPrev!=VAL_NULL) << 1) | (valNext !=VAL_NULL); // binary encoded 00/01/10/11
    // TODO: solve INVALID terminator object problems (example: DLL destructor)
    switch(variant) {
      case 0: // NULL <- SHAPEPROTOTYPE -> NULL
              // O.K. classic NULL-terminated DLL
              shapeType = SHAPE_DLL_NULL;
              break;

      case 1: // NULL <- SHAPEPROTOTYPE -> nonNULL
              // FIXME: extra variants
                return false;       // non-shape
              break;

      case 2: // nonNULL <- SHAPEPROTOTYPE -> NULL
              // FIXME: extra variants (e.g. linux hashlist?)
                return false;       // non-shape
              break;

      case 3: // nonNULL <- SHAPEPROTOTYPE -> nonNULL
            {  // TODO: isolate in extra function
//              CL_WARN("ApparentShapeDetector uses a non-NULL terminator");
              TObjId prev = nextObj(sh_, beg, props.bOff.prev);
              TObjId next = nextObj(sh_, end, props.bOff.next);

              if(!(sh_.isValid(prev) && sh_.isValid(next)))
                  return false;         // use of INVALID object as terminator

              if (valPrev != valNext) { // non-equal end pointers, maybe loop
                  // TODO: how about asymmetric DLL variants? head1/head2, selfloops, ...
                  if(!(prev==end && next==beg))
                      // terminator mismatch
                      return false;
                  // OK, both ends point to each other = cyclic DLL
              shapeType = SHAPE_CYCLIC_DLL;
//                  CL_WARN("ApparentShapeDetector has found cyclic shape");

/////////////////////////////////////////////////////////////////////////////
// NORMALIZE cyclic shape FIXME Warning HACK!
// INFO: cyclic shape can be discovered starting at any object in shape
// we need start the shape at "head", which is (in program always, but not in template)
// pointed by some (program) variable
// This HACK uses minimal ID objects/progvars to fix starting position somewhat
{

    TObjId p = beg;
    TObjId start = beg;
    TObjId min = OBJ_MAX_ID;
    TObjId minObj = beg;
    do {
        if(minObj>p)
            minObj = p;
        //    ETargetSpecifier kind=sh_.objKind(p);
        FldList refs;
        sh_.pointedBy(refs, p);
        BOOST_FOREACH(const FldHandle fld, refs) {
            const TObjId refObj = fld.obj();
            // TODO check if head target ?
            const EStorageClass code = sh_.objStorClass(refObj);
            if (isProgramVar(code)) {
                if (refObj < min) {
                    min = refObj;
                    start = p;
                }
            }
        }
        p = nextObj(sh_, p, props.bOff.next);
    } while (p != beg);

    if (min != OBJ_MAX_ID) {    // found at least one progvar pointing to shape
//std::cout << "i  found sequence start=" << start << "  pointed by progvarID=" << min << std::endl;
#if 0
        beg = nextObj(sh_, start, props.bOff.prev);
#else
        beg = start;            // choose object pointed by progvar with minimal ID number
#endif
    }
    else {                      // no prog var => we are in template
//std::cout << "i  found template sequence start=" << minObj << std::endl;
        beg = minObj;           // choose shape object with minimal ID number
    }

}
/////////////////////////////////////////////////////////////////////////////

                  break;
              }else{ // equal end pointers
                  // the same non-NULL value on both ends, prev==next
                  if(len>1) { // should point to some HEAD
                      // check for pointers inside SHAPEPROTOTYPE
                      BOOST_FOREACH(const TObjId seenObj, seen)
                          if(seenObj==prev)     // FYI: prev==next
                              return false; // TODO: how about self-loop ended DLL ?
                  }
                  // selfloop or some HEAD
                  if(len==1 && beg==next) {
                      // self-loop = single item cyclicDLL
                      shapeType = SHAPE_CYCLIC_DLL;
                      break;
                  }
                  shapeType = SHAPE_CYCLIC_DLL_WITH_HEAD;
                  // TODO: check if HEAD points to SHAPEPROTOTYPE? cDLLwithHEAD
                  // how about non cyclic non-NULL terminated list?
#if 1
                  // this type is sometimes present in middle of cDLL ops =>
                  // FIXME: BUG until shape type is properly checked
                  return false;
#endif
              }
            }
              break;
        default:
            CL_TRAP;
            return false;
    } // switch

//    CS_DEBUG("ApparentShapeDetector found a new container shape, len = " << len);
//std::cout << "i ApparentShapeDetector found a new container shape, len = " << len << std::endl;

    // mark all used objects as taken
    TObjSet &taken = objsByProps_[props];
    if (taken.empty())
        taken.swap(seen);
    else {
        BOOST_FOREACH(const TObjId seenObj, seen)
            taken.insert(seenObj);
    }

    // insert the shape in the destination array
    const Shape shape(
        /* type   */ shapeType,
        /* entry  */ beg,
        /* props  */ props,
        /* length */ len);

//# std::cout << "+++ SHAPE: loc=" << ::debug_location << "/" << ::debug_smg << ", \t " << shape << std::endl;

    dstArray_.push_back(shape);

    // found a new container shape
    return true;
}

void detectApparentShapes(TShapeList &dst, SymHeap &sh)
{
//# std::cout << "[function] detectApparentShapes: loc=" << ::debug_location << ", sh=" << ::debug_smg << std::endl;

    CL_BREAK_IF(!dst.empty());
    ApparentShapeDetector shapeDetector(sh, dst);

    // go through all potential shape container entries
    TObjList heapObjs;
    sh.gatherObjects(heapObjs, isOnHeap);
    BOOST_FOREACH(const TObjId obj, heapObjs) {
        if (sh.objProtoLevel(obj))
            // FIXME: we support only L0 data structures for now
            continue;

        const EObjKind kind = sh.objKind(obj);
        if (OK_DLS == kind) {
            const TSizeRange size = sh.objSize(obj);
            CL_BREAK_IF(!IR::isSingular(size));

            // OK_DLS can be seen as list on its own
            // FIXME: we support only OK_DLS for now
            const ShapeProps dlsProps(OK_DLS,
                    sh.segBinding(obj),
                    size.lo);

            shapeDetector.probeEntry(obj, dlsProps);
            continue;
        }

        TShapePropsList propList;
        digShapePropsCandidates(&propList, sh, obj);

        // go through all potential shape properties candidates
        BOOST_FOREACH(const ShapeProps &props, propList) {
            if (OK_DLS != props.kind)
                // FIXME: we support only OK_DLS for now
                continue;

            shapeDetector.probeEntry(obj, props);
        }
    }
}

typedef std::pair<CVar, TOffset>            TPointer;
typedef std::set<TPointer>                  TPointerSet;
typedef std::set<TPointerSet>               TPointerSetLookup;

struct ShapePattern {
    ShapeProps          props;
    TObjType            type;
    TSizeRange          size;
    TPointerSet         headPtrs;
};

typedef std::vector<ShapePattern>           TShapePatternList;

class ImpliedShapeDetector {
    public:
        ImpliedShapeDetector(bool useHeadChk):
            useHeadChk_(useHeadChk)
        {
        }

        bool indexShape(SymHeap &sh, const Shape &shape);
        void appendImpliedShapes(TShapeList *pDst, SymHeap &sh);

    private:
        const bool                  useHeadChk_;
        TShapePatternList           plist_;
        TPointerSetLookup           index_;
};

void digHeadPtrs(
        TPointerSet                *pDst,
        SymHeap                    &sh,
        const TObjId                beg,
        const TObjId                end,
        const TOffset               offHead)
{
    // gather all references of the boundary objects
    FldList refs;
    sh.pointedBy(refs, beg);
    if (beg != end)
        sh.pointedBy(refs, end);

    // go through the list of references
    BOOST_FOREACH(const FldHandle &fld, refs) {
        const TObjId obj = fld.obj();
        const EStorageClass code = sh.objStorClass(obj);
        if (!isProgramVar(code))
            // not a program variable
            continue;

        const TValId val = fld.value();
        const TOffset off = sh.valOffset(val);
        if (off != offHead)
            // not a head offset
            continue;

        const CVar var = sh.cVarByObject(obj);
        const TPointer ptr(var, fld.offset());
        pDst->insert(ptr);
    }
}

bool ImpliedShapeDetector::indexShape(SymHeap &sh, const Shape &shape)
{
    const TObjId beg = shape.entry;
    const TObjId end = lastObjOfShape(sh, shape);

    if (OK_DLS != shape.props.kind)
        // FIXME: only OK_DLS supported for now
        return false;

    const TValId valPrev = valOfPtr(sh, beg, shape.props.bOff.prev);
    const TValId valNext = valOfPtr(sh, end, shape.props.bOff.next);
    if (VAL_NULL != valPrev || VAL_NULL != valNext)
        // FIXME: only NULL-terminated lists supported for now
        return false;

    TPointerSet headPtrs;
    if (useHeadChk_) {
        const TOffset offHead = shape.props.bOff.head;
        digHeadPtrs(&headPtrs, sh, beg, end, offHead);

        if (!insertOnce(index_, headPtrs))
            // already indexed
            return false;
    }

    ShapePattern sp = {
        /* props    */ shape.props,
        /* type     */ sh.objEstimatedType(beg),
        /* size     */ sh.objSize(beg),
        /* headPtrs */ headPtrs
    };
    plist_.push_back(sp);

    // shape successfully indexed
    return true;
}

bool detectImpliedSingleNode(
        Shape                      *pDst,
        SymHeap                    &sh,
        const ShapePattern         &sp,
        const TValId                valHead)
{
// BUG: head can be on different offset
    const TOffset offHead = sh.valOffset(valHead);
    if (offHead != sp.props.bOff.head)
        // head offset mismatch
        return false;

    const TObjId obj = sh.objByAddr(valHead);
// BUG: if head not on heap
    if (!sh.isValid(obj) || !isOnHeap(sh.objStorClass(obj)))
        // not a valid heap object
        return false;
// TODO FIXME what if head obj not on heap?

// BUG: if head not in the same sized region
    const TSizeRange size = sh.objSize(obj);
    if (size != sp.size)
        // size mismatch
        return false;

// BUG: if head object not the same type as list items
    const TObjType type = sh.objEstimatedType(obj);
    if (type && sp.type && (*type != *sp.type))
        // type mismatch
        return false;

    // prev/next should be NULL for empty list
    const TValId valNext = valOfPtr(sh, obj, sp.props.bOff.next);
    if (VAL_NULL != valNext)
        // next pointer mismatch
        return false;

    const TValId valPrev = valOfPtr(sh, obj, sp.props.bOff.prev);
    if (VAL_NULL != valPrev)
        // prev pointer mismatch
        return false;

    pDst->type   = SHAPE_DLL_NULL;
    pDst->entry  = obj;
    pDst->props  = sp.props;
    pDst->length = 1U;          // do we count head to shape length?

    CS_DEBUG("ImpliedShapeDetector matches a region as container shape");
//# std::cout << "+++ IMPLIED SHAPE1: " << *pDst << std::endl;
    return true;
}

bool detectImpliedShape(Shape *pDst, SymHeap &sh, const ShapePattern &sp)
{
    TValId valHead = VAL_INVALID;

    BOOST_FOREACH(const TPointer &ptr, sp.headPtrs) {
        const CVar &var = ptr.first;
        const TOffset off = ptr.second;

        const TObjId obj = sh.regionByVar(var, /* createIfNeeded */ false);
        if (OBJ_INVALID == obj)
            // missing program variable
            return false;

        const TValId val = valOfPtr(sh, obj, off);
        CL_BREAK_IF(VAL_INVALID == val);

        if (VAL_INVALID == valHead)
            valHead = val;
        else if (val != valHead)
            // head value mismatch
            return false;
    }

    if (VAL_NULL != valHead)
        return detectImpliedSingleNode(pDst, sh, sp, valHead);

#if 0
    // empty list
    pDst->entry  = OBJ_INVALID;
    pDst->props  = sp.props;
    pDst->length = 0U;

    CS_DEBUG("ImpliedShapeDetector matches an empty list as container shape");
    return true;
#else
    return false;
#endif
}

bool detectImpliedShapeBlindly(
        Shape                      *pDst,
        SymHeap                    &sh,
        const ShapePattern         &sp,
        const TObjId                obj)
{
    if (!sh.isValid(obj) || !isOnHeap(sh.objStorClass(obj)))
        // not a valid heap object
        return false;

    const TSizeRange size = sh.objSize(obj);
    if (size != sp.size)
        // size mismatch
        return false;

    const TObjType type = sh.objEstimatedType(obj);
    if (type && sp.type && (*type != *sp.type))
        // type mismatch
        return false;

    const TValId valNext = valOfPtr(sh, obj, sp.props.bOff.next);
    if (VAL_NULL != valNext)
        // next pointer mismatch
        return false;

    const TValId valPrev = valOfPtr(sh, obj, sp.props.bOff.prev);
    if (VAL_NULL != valPrev)
        // prev pointer mismatch
        return false;

    pDst->type   = SHAPE_DLL_NULL;
    pDst->entry  = obj;
    pDst->props  = sp.props;
    pDst->length = 1U;

    CS_DEBUG("ImpliedShapeDetector matches a region as container shape");
//# std::cout << "+++ IMPLIED SHAPE1b: " << *pDst << std::endl;

    return true;
}

void ImpliedShapeDetector::appendImpliedShapes(TShapeList *pDst, SymHeap &sh)
{
    // eliminate duplicates (one container shape can be implied by many)
    TShapeSet found;
    BOOST_FOREACH(const Shape &shape, *pDst)
        found.insert(shape);

unsigned oldsize = pDst->size();
//# std::cout << "[function] ImpliedShapeDetector::appendImpliedShapes initial shapes=" << oldsize << std::endl;

    BOOST_FOREACH(const ShapePattern &sp, plist_) {
        Shape shape(SHAPE_INVALID);

        if (useHeadChk_) {
            if (detectImpliedShape(&shape, sh, sp) && insertOnce(found, shape))
                pDst->push_back(shape);

            continue;
        }

        TObjList allObjs;
        sh.gatherObjects(allObjs);
        BOOST_FOREACH(const TObjId obj, allObjs) {
            if (detectImpliedShapeBlindly(&shape, sh, sp, obj)
                    && insertOnce(found, shape))
                pDst->push_back(shape);
        }
    }

int delta=pDst->size()-oldsize;
if(delta!=0) {
//# std::cout << "+          ImpliedShapeDetector::appendImpliedShapes ADDED shapes=" << delta << std::endl;
}

}

struct DetectionCtx {
    TShapeListByHeapIdx            &dstArray;
    const SymState                 &srcState;
    const unsigned                  cntHeaps;

    DetectionCtx(TShapeListByHeapIdx &dstArray_, const SymState &srcState_):
        dstArray(dstArray_),
        srcState(srcState_),
        cntHeaps(srcState_.size())
    {
    }
};

bool hasAnyProgramVar(const SymState &state)
{
    BOOST_FOREACH(const SymHeap *pSh, state) {
        TObjList vars;
        pSh->gatherObjects(vars, isProgramVar);
        if (!vars.empty())
            return true;
    }

    // found nothing
    return false;
}

void detectImpliedShapes(DetectionCtx &ctx)
{
//# std::cout << "[function] detectImpliedShapes:" << std::endl;

#if 0
    // if there are no variables, we are likely dealing with templates
    const bool anyVars = hasAnyProgramVar(ctx.srcState);
#else
    // FIXME: temporarily disabled in order to find more container shapes
    const bool anyVars = false;
#endif
    ImpliedShapeDetector shapeDetector(/* useHeadChk */ anyVars);

    for (unsigned i = 0U; i < ctx.cntHeaps; ++i) {
        SymHeap &sh = const_cast<SymHeap &>(ctx.srcState[i]);
        const TShapeList &apparentShapes = ctx.dstArray[i];

        BOOST_FOREACH(const Shape &shape, apparentShapes)
            shapeDetector.indexShape(sh, shape);
    }

    for (unsigned i = 0U; i < ctx.cntHeaps; ++i) {
        SymHeap &sh = const_cast<SymHeap &>(ctx.srcState[i]);
        shapeDetector.appendImpliedShapes(&ctx.dstArray[i], sh);
    }
}

} // namespace ContShape

// used for template shape discovery only
void detectLocalContShapes(TShapeListByHeapIdx *pDst, const SymState &state)
{
    CL_BREAK_IF(!pDst->empty());        // output only parameter

    using namespace ContShape;

    const unsigned cnt = state.size();
    pDst->resize(cnt);

    DetectionCtx ctx(*pDst, state);

    // first detect container shapes we can detect on their own
    bool foundApparentShape = false;
    for (unsigned i = 0U; i < cnt; ++i) {
debug_smg = i;  // HACK
        TShapeList &dst = ctx.dstArray[i];
        SymHeap &src = const_cast<SymHeap &>(state[i]);

        detectApparentShapes(dst, src);
        if (dst.empty())
            continue;

        foundApparentShape = true;
    }

    if (!foundApparentShape)
        // no apparent shape found, so we cannot look for implied shapes
        return;

    detectImpliedShapes(ctx);
}
